# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase
from anyblok.registry import RegistryManager
from anyblok.environment import EnvironmentManager


class TestRegistryEntry(TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestRegistryEntry, cls).setUpClass()
        RegistryManager.declare_entry('Other')
        RegistryManager.init_blok('testEntry')
        EnvironmentManager.set('current_blok', 'testEntry')

    @classmethod
    def tearDownClass(cls):
        super(TestRegistryEntry, cls).tearDownClass()
        EnvironmentManager.set('current_blok', None)
        del RegistryManager.loaded_bloks['testEntry']
        RegistryManager.undeclare_entry('Other')

    def tearDown(self):
        super(TestRegistryEntry, self).tearDown()
        del RegistryManager.loaded_bloks['testEntry']['Other']['test']

    def assertInEntry(self, entry, *args):
        blokname = 'testEntry'
        blok = RegistryManager.loaded_bloks[blokname]
        self.assertEqual(len(blok['Other'][entry]['bases']), len(args))
        for cls_ in args:
            hasCls = cls_ in blok['Other'][entry]['bases']
            self.assertEqual(hasCls, True)

    def test_add_entry(self):
        class test:
            pass

        RegistryManager.add_entry_in_register('Other', 'test', test)
        self.assertInEntry('test', test)
        self.assertEqual(
            RegistryManager.has_entry_in_register('testEntry', 'Other',
                                                  'test'),
            True)

    def test_remove_entry(self):
        class test:
            pass

        def has_test_in_removed():
            core = RegistryManager.loaded_bloks['testEntry']['removed']
            if test in core:
                return True

            return False

        RegistryManager.add_entry_in_register('Other', 'test', test)
        self.assertInEntry('test', test)
        self.assertFalse(has_test_in_removed())
        RegistryManager.remove_in_register(test)
        self.assertTrue(has_test_in_removed())

    def test_get_entry_properties_in_register(self):
        class test:
            pass

        RegistryManager.add_entry_in_register('Other', 'test', test,
                                              property1='test')
        properties = RegistryManager.get_entry_properties_in_register(
            'Other', 'test')

        hasproperty1 = 'property1' in properties
        self.assertTrue(hasproperty1)
        self.assertEqual(properties['property1'], 'test')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase, DBTestCase
from sqlalchemy.schema import ForeignKey
from anyblok.mapper import (ModelAttribute, ModelAttributeException,
                            ModelAttributeAdapter, ModelAdapter, ModelRepr,
                            ModelReprException, ModelAttributeAdapterException,
                            ModelMapper, ModelAttributeMapper, MapperAdapter,
                            MapperException)
from anyblok import Declarations
from anyblok.column import String, Integer

register = Declarations.register
unregister = Declarations.unregister
Model = Declarations.Model


class TestModelAttribute(DBTestCase):

    def test_get_attribute(self):
        registry = self.init_registry(None)
        ma = ModelAttribute('Model.System.Model', 'name')
        self.assertEqual(ma.get_attribute(registry),
                         registry.get('Model.System.Model').name)

    def test_without_attribute(self):
        with self.assertRaises(ModelAttributeException):
            ModelAttribute('Model.System.Model', None)

    def test_without_model(self):
        with self.assertRaises(ModelAttributeException):
            ModelAttribute(None, 'name')

    def test_without_model_and_attribute(self):
        with self.assertRaises(ModelAttributeException):
            ModelAttribute(None, None)

    def test_get_attribute_unexisting_model(self):
        registry = self.init_registry(None)
        ma = ModelAttribute('Model.Unexisting.Model', 'name')
        with self.assertRaises(ModelAttributeException):
            ma.get_attribute(registry)

    def test_get_attribute_unexisting_attribute(self):
        registry = self.init_registry(None)
        ma = ModelAttribute('Model.System.Model', 'id')
        with self.assertRaises(ModelAttributeException):
            ma.get_attribute(registry)

    def test_get_fk_name_with_wrong_model(self):
        registry = self.init_registry(None)
        ma = ModelAttribute('Model.System', 'name')
        with self.assertRaises(ModelAttributeException):
            ma.get_fk_name(registry)

    def test_get_fk_name_with_unexisting_model(self):
        registry = self.init_registry(None)
        ma = ModelAttribute('Model.Unexisting.Model', 'name')
        with self.assertRaises(ModelAttributeException):
            ma.get_fk_name(registry)

    def test_get_fk_name_unexisting_attribute(self):
        registry = self.init_registry(None)
        ma = ModelAttribute('Model.System.Model', 'id')
        with self.assertRaises(ModelAttributeException):
            ma.get_fk_name(registry)

    def test_get_fk_name(self):
        registry = self.init_registry(None)
        ma = ModelAttribute('Model.System.Model', 'name')
        self.assertEqual(ma.get_fk_name(registry), 'system_model.name')

    def test_get_fk_name_with_name_different_of_column_name(self):

        def add_in_registry():

            @register(Model)
            class Test:

                name = String(primary_key=True, db_column_name='other')

        registry = self.init_registry(add_in_registry)
        ma = ModelAttribute('Model.Test', 'name')
        self.assertEqual(ma.get_fk_name(registry), 'test.other')

    def test_get_fk(self):
        registry = self.init_registry(None)
        ma = ModelAttribute('Model.System.Model', 'name')
        self.assertTrue(isinstance(ma.get_fk(registry), ForeignKey))

    def test_get_fk_with_options(self):
        registry = self.init_registry(None)
        ma = ModelAttribute('Model.System.Model', 'name').options(
            ondelete='cascade')
        mafk = ma.get_fk(registry)
        self.assertTrue(isinstance(mafk, ForeignKey))

    def test_use(self):
        ma = Declarations.Model.System.Model.use('name')
        self.assertTrue(isinstance(ma, ModelAttribute))
        self.assertEqual(ma.model_name, 'Model.System.Model')
        self.assertEqual(ma.attribute_name, 'name')


class TestModelAttributeAdapter(TestCase):

    def test_from_declaration(self):
        ma = ModelAttribute('Model.System.Model', 'name')
        maa = ModelAttributeAdapter(ma)
        self.assertIs(maa, ma)

    def test_from_registry_name(self):
        maa = ModelAttributeAdapter("Model.System.Model=>name")
        self.assertTrue(isinstance(maa, ModelAttribute))
        self.assertEqual(maa.model_name, 'Model.System.Model')
        self.assertEqual(maa.attribute_name, 'name')

    def test_from_registry_name_without_attribute(self):
        with self.assertRaises(ModelAttributeAdapterException):
            ModelAttributeAdapter("Model.System.Model")


class TestModelRepr(DBTestCase):

    def test_unexisting_model(self):
        registry = self.init_registry(None)
        mr = ModelRepr('Model.Unexisting.Model')
        with self.assertRaises(ModelReprException):
            mr.check_model(registry)

    def test_get_tablename(self):
        registry = self.init_registry(None)
        mr = ModelRepr('Model.System.Model')
        self.assertEqual(mr.tablename(registry), 'system_model')

    def test_get_registry_name(self):
        mr = ModelRepr('Model.System.Model')
        self.assertEqual(mr.model_name, 'Model.System.Model')

    def test_get_primary_keys(self):
        registry = self.init_registry(None)
        mr = ModelRepr('Model.System.Model')
        mas = mr.primary_keys(registry)
        self.assertEqual(len(mas), 1)
        self.assertEqual([x.attribute_name for x in mas], ['name'])

    def test_get_foreign_key_for(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(foreign_key=Model.Test.use('id'))

        registry = self.init_registry(add_in_registry)
        mr = ModelRepr('Model.Test2')
        mas = mr.foreign_keys_for(registry, 'Model.Test')
        self.assertEqual(len(mas), 1)
        self.assertEqual([x.attribute_name for x in mas], ['test_id'])

    def test_without_model(self):
        with self.assertRaises(ModelReprException):
            ModelRepr(None)


class TestModelAdapter(TestCase):

    def test_from_declaration(self):
        mr = ModelRepr('Model.System.Model')
        mra = ModelAdapter(mr)
        self.assertIs(mr, mra)

    def test_from_registry_name(self):
        mra = ModelAdapter("Model.System.Model")
        self.assertTrue(isinstance(mra, ModelRepr))
        self.assertEqual(mra.model_name, 'Model.System.Model')


class TestModelMapper(DBTestCase):

    def test_not_capable(self):
        self.assertFalse(ModelMapper.capable(None))

    def test_capable_by_declaration(self):
        self.assertTrue(ModelMapper.capable(Model.System.Model))

    def test_capable_by_registry_name(self):
        self.assertTrue(ModelMapper.capable('Model.System.Model'))

    def test_by_declaration(self):
        mm = ModelMapper(Model.System.Model, 'even')
        self.assertTrue(isinstance(mm.model, ModelRepr))
        self.assertTrue(mm.model.model_name, 'Model.System.Model')

    def test_by_registry_name(self):
        mm = ModelMapper('Model.System.Model', 'event')
        self.assertTrue(isinstance(mm.model, ModelRepr))
        self.assertTrue(mm.model.model_name, 'Model.System.Model')

    def test_listen_sqlalchemy(self):

        def method():
            pass

        mm = ModelMapper(Model.System.Model, 'before_insert')
        mm.listen(method)
        self.assertTrue(method.is_an_sqlalchemy_event_listener)
        self.assertEqual(method.sqlalchemy_listener, mm)

    def test_listen_anyblok(self):
        def method():
            pass

        mm = ModelMapper(Model.System.Model, 'event')
        mm.listen(method)
        self.assertTrue(method.is_an_event_listener)
        self.assertEqual(method.model, 'Model.System.Model')
        self.assertEqual(method.event, 'event')

    def test_get_mapper(self):
        registry = self.init_registry(None)
        mm = ModelMapper(Model.System.Model, 'even')
        self.assertIs(mm.mapper(registry, None), registry.System.Model)


class TestModelAttributeMapper(DBTestCase):

    def test_not_str_capable(self):
        self.assertFalse(ModelAttributeMapper.capable('Model.System.Model'))

    def test_not_capable(self):
        self.assertFalse(ModelAttributeMapper.capable(None))

    def test_capable_by_declaration(self):
        self.assertTrue(ModelAttributeMapper.capable(
            Model.System.Model.use('name')))

    def test_capable_by_registry_name(self):
        self.assertTrue(ModelAttributeMapper.capable(
            'Model.System.Model=>name'))

    def test_by_declaration(self):
        mam = ModelAttributeMapper(Model.System.Model.use('name'), 'event')
        self.assertTrue(isinstance(mam.attribute, ModelAttribute))
        self.assertTrue(mam.attribute.model_name, 'Model.System.Model')

    def test_by_registry_name(self):
        mam = ModelAttributeMapper('Model.System.Model=>name', 'event')
        self.assertTrue(isinstance(mam.attribute, ModelAttribute))
        self.assertTrue(mam.attribute.model_name, 'Model.System.Model')

    def test_listen(self):

        def method():
            pass

        mam = ModelAttributeMapper(Model.System.Model.use('name'), 'set')
        mam.listen(method)
        self.assertTrue(method.is_an_sqlalchemy_event_listener)
        self.assertEqual(method.sqlalchemy_listener, mam)

    def test_get_mapper(self):
        registry = self.init_registry(None)
        mam = ModelAttributeMapper(Model.System.Model.use('name'), 'set')
        # We can't compare that the column are the same, because is the case
        # of the call are in the class attribute (no instance) SQLAlchemy
        # Wrap the result for each call, then each call return a différent
        # object, but it is not an error
        self.assertEqual(
            str(mam.mapper(registry, None) == 'test'),
            str(registry.System.Model.name == 'test'))


class TestMapperAdapter(TestCase):

    def test_model_mapper(self):
        mam = MapperAdapter('Model.System.Model', 'event')
        self.assertTrue(isinstance(mam, ModelMapper))

    def test_model_attribute_mapper(self):
        mam = MapperAdapter('Model.System.Model=>name', 'event')
        self.assertTrue(isinstance(mam, ModelAttributeMapper))

    def test_no_mapper(self):
        with self.assertRaises(MapperException):
            MapperAdapter(None)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase


class TestSystemBlok(BlokTestCase):

    def test_list_by_state_installed(self):
        installed = self.registry.System.Blok.list_by_state('installed')
        core_is_installed = 'anyblok-core' in installed
        self.assertEqual(core_is_installed, True)

    def test_list_by_state_without_state(self):
        self.assertEqual(self.registry.System.Blok.list_by_state(), None)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .plugins import ModelPluginBase
from anyblok.common import apply_cache


class CachePlugin(ModelPluginBase):

    def __init__(self, registry):
        if not hasattr(registry, 'caches'):
            registry.caches = {}

        super(CachePlugin, self).__init__(registry)

    def transform_base_attribute(self, attr, method, namespace, base,
                                 transformation_properties,
                                 new_type_properties):
        """ Find the sqlalchemy hybrid methods in the base to save the
        namespace and the method in the registry

        :param attr: attribute name
        :param method: method pointer of the attribute
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param new_type_properties: param to add in a new base if need
        """
        new_type_properties.update(apply_cache(
            attr, method, self.registry, namespace, base,
            transformation_properties))



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase, DBTestCase
from anyblok.field import FieldException
from anyblok.relationship import RelationShip
from anyblok import Declarations
from anyblok.column import Integer
from anyblok.relationship import One2Many, Many2One


register = Declarations.register
Model = Declarations.Model


class OneRelationShip(RelationShip):
    pass


class OneModel:
    __tablename__ = 'test'
    __registry_name__ = 'One.Model'


class MockRegistry:

    InstrumentedList = []


class TestRelationShip(TestCase):

    def test_forbid_instance(self):
        try:
            RelationShip(model=OneModel)
            self.fail("RelationShip mustn't be instanciated")
        except FieldException:
            pass

    def test_must_have_a_model(self):
        OneRelationShip(model=OneModel)
        try:
            OneRelationShip()
            self.fail("No watchdog, the model must be required")
        except FieldException:
            pass


class TestComplexeRelationShipCase(DBTestCase):

    def test_with_multi_parallel_foreign_key_with_definition(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                id = Integer(primary_key=True)
                address_1_id = Integer(foreign_key=Model.Address.use('id'))
                address_2_id = Integer(foreign_key=Model.Address.use('id'))
                address_1 = Many2One(model=Model.Address,
                                     column_names=['address_1_id'])
                address_2 = Many2One(model=Model.Address,
                                     column_names=['address_2_id'])

            @register(Model)  # noqa
            class Address:

                persons = One2Many(model=Model.Person)

        registry = self.init_registry(add_in_registry)
        address_1 = registry.Address.insert()
        address_2 = registry.Address.insert()
        person = registry.Person.insert(
            address_1=address_1, address_2=address_2)
        self.assertEqual(address_1.persons, [person])
        self.assertEqual(address_2.persons, [person])

    def test_with_multi_parallel_foreign_key_auto_detect(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                persons = One2Many(model='Model.Person')

            @register(Model)
            class Person:

                id = Integer(primary_key=True)
                address_1 = Many2One(model=Model.Address)
                address_2 = Many2One(model=Model.Address)

        registry = self.init_registry(add_in_registry)
        address_1 = registry.Address.insert()
        address_2 = registry.Address.insert()
        person = registry.Person.insert(
            address_1=address_1, address_2=address_2)
        self.assertEqual(address_1.persons, [person])
        self.assertEqual(address_2.persons, [person])



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok import Declarations
from sqlalchemy.exc import IntegrityError
from anyblok.field import FieldException
from anyblok.column import (
    Integer, String, BigInteger, Float, Decimal, Boolean, DateTime, Date, Time,
    Sequence)
from anyblok.relationship import Many2One
from sqlalchemy import ForeignKeyConstraint
from sqlalchemy.engine.reflection import Inspector


register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin


def _complete_many2one(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address,
                           remote_columns="id", column_names="id_of_address",
                           one2many="persons", nullable=False)


def _many2one_with_same_name_for_column_names(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address,
                           column_names="address")


def _minimum_many2one(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address)


def _many2one_with_str_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model='Model.Address')


def _minimum_many2one_without_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One()


def _auto_detect_type(ColumnType=None, **kwargs):

    @register(Model)
    class Address:

        id = ColumnType(primary_key=True, **kwargs)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = Many2One(model=Model.Address)


def _two_remote_primary_keys(**kwargs):

    @register(Model)
    class Test:

        id = Integer(primary_key=True, unique=True)
        id2 = String(primary_key=True, unique=True)

    @register(Model)
    class Test2:

        id = Integer(primary_key=True)
        test = Many2One(model=Model.Test)


class TestMany2One(DBTestCase):

    def test_complete_many2one(self):
        registry = self.init_registry(_complete_many2one)
        address_exist = hasattr(registry.Person, 'address')
        self.assertTrue(address_exist)
        id_of_address_exist = hasattr(registry.Person, 'id_of_address')
        self.assertTrue(id_of_address_exist)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        self.assertEqual(address.persons, [person])

    def test_complete_many2one_expire_field(self):
        registry = self.init_registry(_complete_many2one)
        self.assertIn(
            ('address',),
            registry.expire_attributes['Model.Person']['id_of_address'])

    def test_2_many2one(self):
        with self.assertRaises(FieldException):
            self.init_registry(_many2one_with_same_name_for_column_names)

    def test_minimum_many2one(self):
        registry = self.init_registry(_minimum_many2one)
        address_exist = hasattr(registry.Person, 'address')
        self.assertTrue(address_exist)
        address_id_exist = hasattr(registry.Person, 'address_id')
        self.assertTrue(address_id_exist)
        address = registry.Address.insert()
        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)
        self.assertEqual(person.address, address)

    def test_minimum_many2one_on_sequence(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                seq = Sequence(primary_key=True, formater="V-{seq}")

            @register(Model)
            class Test2:

                seq = Sequence(primary_key=True)
                test = Many2One(model=Model.Test)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        test2 = registry.Test2.insert(test=test)
        self.assertIs(test, test2.test)

    def test_minimum_many2one_expire_field(self):
        registry = self.init_registry(_minimum_many2one)
        self.assertEqual(
            registry.expire_attributes['Model.Person']['address_id'],
            {('address',)})

    def test_many2one_with_str_model(self):
        registry = self.init_registry(_many2one_with_str_model)
        address_exist = hasattr(registry.Person, 'address')
        self.assertTrue(address_exist)
        address_id_exist = hasattr(registry.Person, 'address_id')
        self.assertTrue(address_id_exist)
        address = registry.Address.insert()
        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)
        self.assertIs(person.address, address)

    def test_minimum_many2one_without_model(self):
        with self.assertRaises(FieldException):
            self.init_registry(_minimum_many2one_without_model)

    def check_autodetect_type(self, ColumnType):
        registry = self.init_registry(_auto_detect_type, ColumnType=ColumnType)
        self.assertEqual(
            str(registry.Address.id.property.columns[0].type),
            str(registry.Person.address_id.property.columns[0].type))

    def test_autodetect_type_integer(self):
        self.check_autodetect_type(Integer)

    def test_autodetect_type_big_integer(self):
        self.check_autodetect_type(BigInteger)

    def test_autodetect_type_float(self):
        self.check_autodetect_type(Float)

    def test_autodetect_type_decimal(self):
        self.check_autodetect_type(Decimal)

    def test_autodetect_type_string(self):
        self.check_autodetect_type(String)

    def test_autodetect_type_boolean(self):
        self.check_autodetect_type(Boolean)

    def test_autodetect_type_datetime(self):
        self.check_autodetect_type(DateTime)

    def test_autodetect_type_date(self):
        self.check_autodetect_type(Date)

    def test_autodetect_type_time(self):
        self.check_autodetect_type(Time)

    def test_same_model(self):
        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                parent = Many2One(model='Model.Test', one2many='children')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        self.assertEqual(t2.parent, t1)
        self.assertEqual(t1.children[0], t2)

    def test_same_model_pk_by_inherit(self):
        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)  # noqa
            class Test:

                parent = Many2One(model='Model.Test', one2many='children')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        self.assertEqual(t2.parent, t1)
        self.assertEqual(t1.children[0], t2)

    def test_same_model_pk_by_mixin(self):
        def add_in_registry():

            @register(Mixin)
            class MTest:

                id = Integer(primary_key=True)

            @register(Model)  # noqa
            class Test(Mixin.MTest):

                parent = Many2One(model='Model.Test', one2many='children')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        self.assertEqual(t2.parent, t1)
        self.assertEqual(t1.children[0], t2)

    def test_add_unique_constraint(self):
        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address = Many2One(model=Model.Address, unique=True)

        registry = self.init_registry(add_in_registry)
        address = registry.Address.insert()
        registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        with self.assertRaises(IntegrityError):
            registry.Person.insert(name="Other", address=address)

    def test_add_index_constraint(self):
        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address = Many2One(model=Model.Address, index=True)

        registry = self.init_registry(add_in_registry)
        inspector = Inspector(registry.session.connection())
        indexes = inspector.get_indexes(registry.Person.__tablename__)
        self.assertEqual(len(indexes), 1)

    def test_add_primary_keys_constraint(self):
        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address = Many2One(model=Model.Address, primary_key=True)

        registry = self.init_registry(add_in_registry)
        inspector = Inspector(registry.session.connection())
        pks = inspector.get_primary_keys(registry.Person.__tablename__)
        self.assertIn('address_id', pks)

    def test_complet_with_multi_foreign_key(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                @classmethod
                def define_table_args(cls):
                    table_args = super(Test2, cls).define_table_args()
                    return table_args + (ForeignKeyConstraint(
                        [cls.test_id, cls.test_id2], ['test.id', 'test.id2']),)

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test,
                                remote_columns=('id', 'id2'),
                                column_names=('test_id', 'test_id2'))

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_complet_with_multi_foreign_key_without_constraint(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test,
                                remote_columns=('id', 'id2'),
                                column_names=('test_id', 'test_id2'))

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_with_multi_foreign_key_on_existing_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_with_multi_foreign_key_on_existing_column2(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                other_test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                other_test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = Many2One(model=Model.Test)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.other_test_id)
        self.assertEqual(test.id2, test2.other_test_id2)

    def test_with_multi_foreign_key_on_unexisting_column(self):
        registry = self.init_registry(_two_remote_primary_keys)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_many2one_with_multi_fk_expire_field(self):
        registry = self.init_registry(_two_remote_primary_keys)
        self.assertIn(
            ('test',),
            registry.expire_attributes['Model.Test2']['test_id'])
        self.assertIn(
            ('test',),
            registry.expire_attributes['Model.Test2']['test_id2'])

    def test_with_multi_foreign_key_on_unexisting_named_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test,
                                column_names=('test_id', 'test_id2'))

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_with_multi_foreign_key_on_unexisting_named_column2(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = Integer(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, column_names=(
                    'other_test_id', 'other_test_id2'))

        with self.assertRaises(FieldException):
            self.init_registry(add_in_registry)

    def test_m2o_in_mixin(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Mixin)
            class MTest:

                test = Many2One(model=Model.Test)

            @register(Model)
            class Test1(Mixin.MTest):

                id = Integer(primary_key=True)

            @register(Model)
            class Test2(Mixin.MTest):

                id = Integer(primary_key=True)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        test1 = registry.Test1.insert(test=test)
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test1.test_id, test2.test_id)

    def test_delete_m2o_without_fk_options_on_delete_cascade(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)
            class TestM2O:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, nullable=True)

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        testM2O = registry.TestM2O.insert(test=test)
        self.assertIs(testM2O.test, test)
        with self.assertRaises(IntegrityError):
            test.delete()

    def test_delete_m2o_with_fk_options_on_delete_cascade(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)

            @register(Model)
            class TestM2O:

                id = Integer(primary_key=True)
                test = Many2One(model=Model.Test, nullable=True,
                                foreign_key_options={'ondelete': 'cascade'})

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert()
        testM2O = registry.TestM2O.insert(test=test)
        self.assertIs(testM2O.test, test)
        test.delete()
        self.assertFalse(registry.TestM2O.query().count())

    def test_2_simple_many2one_on_the_same_model(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address_1 = Many2One(model=Model.Address)
                address_2 = Many2One(model=Model.Address)

        registry = self.init_registry(add_in_registry)
        address_1 = registry.Address.insert()
        address_2 = registry.Address.insert()
        person = registry.Person.insert(
            name='test', address_1=address_1, address_2=address_2)
        self.assertIs(person.address_1, address_1)
        self.assertIs(person.address_2, address_2)
        self.assertIsNot(person.address_1, person.address_2)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from os.path import join
from anyblok import config
from anyblok.config import (
    Configuration,
    add_configuration_file,
    add_plugins,
    add_database,
    add_install_bloks,
    add_uninstall_bloks,
    add_update_bloks,
    add_interpreter,
    add_schema,
    add_doc,
    add_unittest,
    ConfigurationException,
    AnyBlokActionsContainer,
    ConfigOption,
    AnyBlokPlugin,
    define_preload_option,
)
from anyblok.tests.testcase import TestCase
from sqlalchemy.engine.url import make_url
from anyblok.config import get_url


old_getParser = config.getParser


def fnct_configuration(parser, default):
    default.update({'test': None})


def fnct_other_configuration(parser, default):
    default.update({'test': None})


def MockPluginFnct():
    pass


class MockPluginClass:
    pass


class MockArgParseArguments:

    def __init__(self, configfile=None, args=None, kwargs=None):
        if configfile:
            cfile = join('/'.join(__file__.split('/')[:-1]), configfile)
            self.configfile = cfile
            if kwargs is None:
                kwargs = {}

            kwargs['configfile'] = configfile
        else:
            self.configfile = None
        self.args = args
        self.kwargs = kwargs
        self.logging_level = None
        self.logging_configfile = None
        self.json_logging_configfile = None
        self.yaml_logging_configfile = None

    def _get_args(self):
        return self.args

    def _get_kwargs(self):
        if self.kwargs is None:
            return tuple()
        return self.kwargs.items()


class MockArguments:

    configfile = None
    logging_level = None
    logging_configfile = None
    json_logging_configfile = None
    yaml_logging_configfile = None
    vals = {'var1': 'val1',
            'var2': 'val2',
            'var3': 'val3'}

    def _get_kwargs(self):
        return self.vals.items()

    def _get_args(self):
        return False


class MockArgumentValue:

    def __init__(self, **kwargs):
        self.default = None
        self.type = str
        self.__dict__.update(kwargs)


class MockArgumentParser:

    def __init__(self, *args, **kwargs):
        self.args = MockArguments()

    def add_argument_group(self, *args, **kwargs):
        return MockArgumentParser()

    def parse_args(self, *args, **kwargs):
        return self.args

    def add_argument(self, *args, **kwargs):
        return MockArgumentValue(**kwargs)

    def set_defaults(self, *args, **kwargs):
        pass


class TestConfiguration(TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestConfiguration, cls).setUpClass()

        def getParser(*args, **kwargs):
            return MockArgumentParser()

        config.getParser = getParser
        cls.old_configuration = Configuration.configuration.copy()
        cls.old_groups = Configuration.groups.copy()
        cls.old_labels = Configuration.labels.copy()

    @classmethod
    def tearDownClass(cls):
        super(TestConfiguration, cls).tearDownClass()
        config.getParser = old_getParser
        Configuration.configuration = cls.old_configuration
        Configuration.groups = cls.old_groups
        Configuration.labels = cls.old_labels

    def setUp(self):
        super(TestConfiguration, self).setUp()
        Configuration.groups = {}
        Configuration.labels = {}
        Configuration.configuration = {}

    def assertAdded(self, group, label=None, function_=None):
        self.assertEqual(Configuration.groups[group], [function_])

        if label:
            self.assertEqual(Configuration.labels[group], label)

    def test_add(self):
        Configuration.add('new-group', function_=fnct_configuration)
        self.assertAdded('new-group', function_=fnct_configuration)

    def test_add_with_label(self):
        Configuration.add(
            'new-group', label="One label", function_=fnct_configuration)
        self.assertAdded(
            'new-group', label="One label", function_=fnct_configuration)

    def test_add_decorator(self):

        @Configuration.add('new-group')
        def fnct(parser, default):
            default.update({'other test': True})

        self.assertAdded('new-group', function_=fnct)

    def test_has(self):
        self.assertFalse(Configuration.has('option'))
        Configuration.configuration['option'] = ConfigOption('option', str)
        self.assertTrue(Configuration.has('option'))

    def test_get(self):
        option = 'My option'
        Configuration.configuration['option'] = ConfigOption(option, str)
        res = Configuration.get('option')
        self.assertEqual(option, res)

    def test_fnct_plugins_config(self):
        option = 'anyblok.tests.test_config:MockPluginFnct'
        Configuration.configuration['option'] = ConfigOption(
            option, AnyBlokPlugin)
        res = Configuration.get('option')
        self.assertIs(MockPluginFnct, res)

    def test_class_plugins_config(self):
        option = 'anyblok.tests.test_config:MockPluginClass'
        Configuration.configuration['option'] = ConfigOption(
            option, AnyBlokPlugin)
        res = Configuration.get('option')
        self.assertIs(MockPluginClass, res)

    def test_wrong_plugins_config(self):
        option = 'anyblok.tests.test_config:MockPluginWrong'
        with self.assertRaises(ImportError):
            Configuration.configuration['option'] = ConfigOption(
                option, AnyBlokPlugin)

    def test_update(self):
        Configuration.update(one_option=1)
        self.assertEqual(Configuration.get('one_option'), 1)

    def test_update2(self):
        Configuration.update(dict(one_option=1))
        self.assertEqual(Configuration.get('one_option'), 1)

    def test_set_str(self):
        Configuration.set('value', '1')
        self.assertEqual(Configuration.configuration['value'].type, str)
        self.assertEqual(Configuration.get('value'), '1')

    def test_set_int(self):
        Configuration.set('value', 1)
        self.assertEqual(Configuration.configuration['value'].type, int)
        self.assertEqual(Configuration.get('value'), 1)

    def test_set_float(self):
        Configuration.set('value', 1.)
        self.assertEqual(Configuration.configuration['value'].type, float)
        self.assertEqual(Configuration.get('value'), 1.)

    def test_set_list(self):
        Configuration.set('value', [1])
        self.assertEqual(Configuration.configuration['value'].type, list)
        self.assertEqual(Configuration.get('value'), [1])

    def test_set_tuple(self):
        Configuration.set('value', (1,))
        self.assertEqual(Configuration.configuration['value'].type, tuple)
        self.assertEqual(Configuration.get('value'), (1,))

    def test_set_dict(self):
        Configuration.set('value', {'a': 1})
        self.assertEqual(Configuration.configuration['value'].type, dict)
        self.assertEqual(Configuration.get('value'), {'a': 1})

    def test_get_use_default_value(self):
        option = 'My option by default'
        res = Configuration.get('option', option)
        self.assertEqual(option, res)

    def check_url(self, url, wanted_url):
        wanted_url = make_url(wanted_url)
        for x in ('drivername', 'host', 'port', 'username', 'password',
                  'database'):
            self.assertEqual(
                getattr(url, x), getattr(wanted_url, x),
                "check url(%s) == url(%s) on attribute %r" % (url, wanted_url,
                                                              x))

    def test_get_url(self):
        with TestCase.Configuration(db_name='anyblok',
                                    db_driver_name='postgres',
                                    db_host='localhost',
                                    db_user_name=None,
                                    db_password=None,
                                    db_port=None):
            url = get_url()
            self.check_url(url, 'postgres://localhost/anyblok')

    def test_get_url2(self):
        with TestCase.Configuration(db_name='anyblok',
                                    db_driver_name='postgres',
                                    db_host='localhost',
                                    db_user_name=None,
                                    db_password=None,
                                    db_port=None):
            url = get_url(db_name='anyblok2')
            self.check_url(url, 'postgres://localhost/anyblok2')

    def test_get_url3(self):
        with TestCase.Configuration(db_url='postgres:///anyblok',
                                    db_name=None,
                                    db_driver_name=None,
                                    db_host=None,
                                    db_user_name=None,
                                    db_password=None,
                                    db_port=None):
            url = get_url()
            self.check_url(url, 'postgres:///anyblok')

    def test_get_url4(self):
        with TestCase.Configuration(db_url='postgres:///anyblok',
                                    db_name='anyblok2',
                                    db_driver_name=None,
                                    db_host=None,
                                    db_user_name='jssuzanne',
                                    db_password='secret',
                                    db_port=None):
            url = get_url()
            self.check_url(url, 'postgres://jssuzanne:secret@/anyblok2')

    def test_get_url5(self):
        with TestCase.Configuration(db_url='postgres:///anyblok',
                                    db_name='anyblok2',
                                    db_driver_name=None,
                                    db_host=None,
                                    db_user_name='jssuzanne',
                                    db_password='secret',
                                    db_port=None):
            url = get_url(db_name='anyblok3')
            self.check_url(url, 'postgres://jssuzanne:secret@/anyblok3')

    def test_get_url_without_drivername(self):
        with TestCase.Configuration(db_name=None,
                                    db_driver_name=None,
                                    db_host=None,
                                    db_user_name=None,
                                    db_password=None,
                                    db_port=None):
            with self.assertRaises(ConfigurationException):
                get_url()

    def test_remove(self):
        Configuration.add('new-group', function_=fnct_configuration)
        Configuration.remove('new-group', function_=fnct_configuration)
        self.assertEqual(Configuration.groups['new-group'], [])

    def test_remove_more_function(self):
        Configuration.add('new-group', function_=fnct_configuration)
        Configuration.add('new-group', function_=fnct_other_configuration)
        Configuration.remove('new-group', function_=fnct_configuration)
        self.assertEqual(Configuration.groups['new-group'],
                         [fnct_other_configuration])

    def test_remove_label(self):
        Configuration.add(
            'new-group', label="One label", function_=fnct_configuration)
        Configuration.remove_label('new-group')
        with self.assertRaises(KeyError):
            Configuration.labels['AnyBlok']['new-group']

    def test_load_without_configuration_groupes(self):
        self.assertEqual(Configuration.load('default'), None)

    def test_empty_parse_option(self):
        args = MockArgParseArguments()
        Configuration.parse_options(args)
        self.assertEqual(Configuration.configuration, {})

    def assertConfig(self, kwargs):
        config = {x: y.get() for x, y in Configuration.configuration.items()}
        self.assertEqual(config, kwargs)

    def test_parse_option(self):
        kwargs = {'test': 'value'}
        args = MockArgParseArguments(configfile="mock_configuration_file.cfg",
                                     kwargs=kwargs)
        Configuration.parse_options(args)
        kwargs.update({
            'db_name': 'anyblok',
            'db_driver_name': 'postgres',
            'db_user_name': '',
            'db_password': '',
            'db_host': 'localhost',
            'db_port': '',
        })
        self.assertConfig(kwargs)

    def test_parse_option_configuration(self):
        args = MockArgParseArguments(configfile="mock_configuration_file.cfg")
        Configuration.parse_options(args)
        self.assertConfig({
            'db_name': 'anyblok',
            'db_driver_name': 'postgres',
            'db_user_name': '',
            'db_password': '',
            'db_host': 'localhost',
            'db_port': '',
            'configfile': 'mock_configuration_file.cfg',
        })

    def test_parse_option_configuration_with_extend(self):
        args = MockArgParseArguments(
            configfile="mockblok/mock_configuration_file.cfg")
        Configuration.parse_options(args)
        self.assertConfig({
            'db_name': 'anyblok',
            'db_driver_name': 'postgres',
            'db_user_name': '',
            'db_password': '',
            'db_host': 'localhost',
            'db_port': '',
            'configfile': 'mockblok/mock_configuration_file.cfg',
        })

    def test_parse_option_kwargs(self):
        kwargs = {'test': 'value'}
        args = MockArgParseArguments(kwargs=kwargs)
        Configuration.parse_options(args)
        self.assertConfig(kwargs)

    def test_parse_option_args(self):
        args = ('test',)
        args = MockArgParseArguments(args=args)
        with self.assertRaises(ConfigurationException):
            Configuration.parse_options(args)

    def test_load_with_configuration_groupes(self):
        Configuration.load('default', configuration_groups=['install-bloks'])
        self.assertConfig(MockArguments.vals)

    def test_load_with_bad_configuration_groupes(self):
        Configuration.load('default', configuration_groups=['bad-groups'])
        self.assertConfig(MockArguments.vals)

    def get_parer(self):

        class Parser(AnyBlokActionsContainer, MockArgumentParser):
            pass

        return Parser()

    def test_add_argument_str(self):
        parser = self.get_parer()
        parser.add_argument('--value', dest='value', default='1')
        self.assertEqual(Configuration.configuration['value'].type, str)
        self.assertEqual(Configuration.get('value'), '1')

    def test_add_argument_int(self):
        parser = self.get_parer()
        parser.add_argument('--value', dest='value', type=int, default=1)
        self.assertEqual(Configuration.configuration['value'].type, int)
        self.assertEqual(Configuration.get('value'), 1)

    def test_add_argument_float(self):
        parser = self.get_parer()
        parser.add_argument('--value', dest='value', type=float, default=1)
        self.assertEqual(Configuration.configuration['value'].type, float)
        self.assertEqual(Configuration.get('value'), 1.)

    def test_add_argument_list(self):
        parser = self.get_parer()
        parser.add_argument('--value', dest='value', nargs="+", default='1, 2')
        self.assertEqual(Configuration.get('value'), ['1', '2'])

    def test_default_str(self):
        parser = self.get_parer()
        parser.add_argument('--value', dest='value', default='1')
        parser.set_defaults(value='2')
        self.assertEqual(Configuration.get('value'), '2')


class TestConfigurationOption(TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestConfigurationOption, cls).setUpClass()
        cls.parser = MockArgumentParser()
        cls.group = cls.parser.add_argument_group('label')
        cls.function = {
            'add_configuration_file': add_configuration_file,
            'add_plugins': add_plugins,
            'add_database': add_database,
            'add_install_bloks': add_install_bloks,
            'add_uninstall_bloks': add_uninstall_bloks,
            'add_update_bloks': add_update_bloks,
            'add_interpreter': add_interpreter,
            'add_schema': add_schema,
            'add_doc': add_doc,
            'add_unittest': add_unittest,
            'define_preload_option': define_preload_option,
        }

    def test_add_configuration_file(self):
        self.function['add_configuration_file'](self.parser)

    def test_add_plugins(self):
        self.function['add_plugins'](self.group)

    def test_add_database(self):
        self.function['add_database'](self.group)

    def test_add_install_bloks(self):
        self.function['add_install_bloks'](self.parser)

    def test_add_uninstall_bloks(self):
        self.function['add_uninstall_bloks'](self.parser)

    def test_add_update_bloks(self):
        self.function['add_update_bloks'](self.parser)

    def test_add_interpreter(self):
        self.function['add_interpreter'](self.parser)

    def test_add_schema(self):
        self.function['add_schema'](self.group)

    def test_add_doc(self):
        self.function['add_doc'](self.group)

    def test_add_unittest(self):
        self.function['add_unittest'](self.group)

    def test_define_preload_option(self):
        self.function['define_preload_option'](self.parser)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok import Declarations
from anyblok.field import FieldException
from anyblok.column import Integer, String
from anyblok.relationship import One2One
from sqlalchemy import ForeignKeyConstraint

register = Declarations.register
Model = Declarations.Model


def _complete_one2one(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model=Model.Address,
                          remote_columns="id", column_names="id_of_address",
                          backref="person", nullable=False)


def _minimum_one2one(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model=Model.Address, backref="person")


def _one2one_with_str_method(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model="Model.Address", backref="person")


def _minimum_one2one_without_backref(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model=Model.Address)


def _minimum_one2one_with_one2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model=Model.Address,
                          backref="person", one2many="persons")


def _multi_fk_one2one():

    @register(Model)
    class Test:

        id = Integer(primary_key=True, unique=True)
        id2 = String(primary_key=True, unique=True)

    @register(Model)
    class Test2:

        id = Integer(primary_key=True)
        test = One2One(model=Model.Test,
                       column_names=('test_id', 'test_id2'),
                       backref="test2")


class TestOne2One(DBTestCase):

    def test_complete_one2one(self):
        registry = self.init_registry(_complete_one2one)
        self.assertTrue(hasattr(registry.Person, 'address'))
        self.assertTrue(hasattr(registry.Person, 'id_of_address'))
        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')
        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        self.assertIs(address.person, person)

    def test_minimum_one2one(self):
        registry = self.init_registry(_minimum_one2one)
        self.assertTrue(hasattr(registry.Person, 'address'))
        self.assertTrue(hasattr(registry.Person, 'address_id'))
        address = registry.Address.insert()
        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        self.assertIs(address.person, person)

    def test_one2one_multi_entry_for_same(self):
        registry = self.init_registry(_minimum_one2one)
        address = registry.Address.insert()
        p1 = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)
        self.assertIs(p1.address, address)
        p2 = registry.Person.insert(name="Franck BRET", address=address)
        self.assertIs(p2.address, address)
        self.assertIsNone(p1.address_id)

    def test_minimum_one2one_expire_field(self):
        registry = self.init_registry(_minimum_one2one)
        self.assertIn(
            ('address',),
            registry.expire_attributes['Model.Person']['address_id'])
        self.assertIn(
            ('address', 'person'),
            registry.expire_attributes['Model.Person']['address_id'])

    def test_one2one_with_str_model(self):
        registry = self.init_registry(_one2one_with_str_method)
        self.assertTrue(hasattr(registry.Person, 'address'))
        self.assertTrue(hasattr(registry.Person, 'address_id'))
        address = registry.Address.insert()
        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)
        self.assertIs(address.person, person)

    def test_minimum_one2one_without_backref(self):
        with self.assertRaises(FieldException):
            self.init_registry(_minimum_one2one_without_backref)

    def test_minimum_one2one_with_one2many(self):
        with self.assertRaises(FieldException):
            self.init_registry(_minimum_one2one_with_one2many)

    def test_complet_with_multi_foreign_key(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                @classmethod
                def define_table_args(cls):
                    table_args = super(Test2, cls).define_table_args()
                    return table_args + (ForeignKeyConstraint(
                        [cls.test_id, cls.test_id2], ['test.id', 'test.id2']),)

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = One2One(model=Model.Test,
                               remote_columns=('id', 'id2'),
                               column_names=('test_id', 'test_id2'),
                               backref="test2")

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_complet_with_multi_foreign_key_without_constraint(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = One2One(model=Model.Test,
                               remote_columns=('id', 'id2'),
                               column_names=('test_id', 'test_id2'),
                               backref="test2")

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_with_multi_foreign_key_on_existing_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = One2One(model=Model.Test, backref="test2")

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_with_multi_foreign_key_on_existing_column2(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                other_test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                other_test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = One2One(model=Model.Test, backref="test2")

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.other_test_id)
        self.assertEqual(test.id2, test2.other_test_id2)

    def test_with_multi_foreign_key_on_unexisting_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test = One2One(model=Model.Test, backref="test2")

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_with_multi_foreign_key_on_unexisting_named_column(self):
        registry = self.init_registry(_multi_fk_one2one)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        self.assertEqual(test.id, test2.test_id)
        self.assertEqual(test.id2, test2.test_id2)

    def test_multi_o2o_one2one_expire_field(self):
        registry = self.init_registry(_multi_fk_one2one)
        self.assertIn(
            ('test',),
            registry.expire_attributes['Model.Test2']['test_id'])
        self.assertIn(
            ('test', 'test2'),
            registry.expire_attributes['Model.Test2']['test_id'])
        self.assertIn(
            ('test',),
            registry.expire_attributes['Model.Test2']['test_id2'])
        self.assertIn(
            ('test', 'test2'),
            registry.expire_attributes['Model.Test2']['test_id2'])

    def test_with_multi_foreign_key_on_unexisting_named_column2(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = Integer(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test = One2One(model=Model.Test, column_names=(
                    'other_test_id', 'other_test_id2'), backref="test2")

        with self.assertRaises(FieldException):
            self.init_registry(add_in_registry)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import anyblok
from anyblok.release import version
from anyblok.blok import BlokManager
from anyblok.config import Configuration, get_db_name
from anyblok.registry import RegistryManager, return_list
from anyblok._graphviz import ModelSchema, SQLSchema
from nose import main
import sys
from os.path import join
from os import walk
from argparse import RawDescriptionHelpFormatter
from textwrap import dedent
from sqlalchemy_utils.functions import create_database
from anyblok import (
    load_init_function_from_entry_points,
    configuration_post_load,
)

Configuration.add_application_properties(
    'createdb',
    [
        'unittest',
        'logging',
        'create_db',
        'install-bloks',
        'install-or-update-bloks',
    ],
    prog='AnyBlok create database, version %r' % version,
    description="Create a database and install bloks to populate it"
)

Configuration.add_application_properties(
    'updatedb',
    [
        'unittest',
        'logging',
        'install-bloks',
        'uninstall-bloks',
        'update-bloks',
        'install-or-update-bloks',
    ],
    prog='AnyBlok update database, version %r' % version,
    description="Update a database: install, upgrade or uninstall the bloks "
)

Configuration.add_application_properties(
    'nose', ['logging', 'unittest'],
    prog='AnyBlok nose, version %r' % version,
    description="Run fonctionnal nosetest of the installed bloks"
)

Configuration.add_application_properties(
    'interpreter', ['logging', 'interpreter'],
    prog='AnyBlok interpretor, version %r' % version,
    description="Run an interpreter on the registry",
    formatter_class=RawDescriptionHelpFormatter,
    epilog=dedent("Example\n"
                  "-------\n"
                  "  $ anyblok_interpreter [anyblok arguments] \n"
                  "  $ => registry \n"
                  "  ... <registry> \n\n"
                  "  The interpretor add in the local the registry of "
                  "the selected database \n\n"
                  "Note\n"
                  "----\n"
                  "  if the ipython is in the python path, then "
                  "the interpretor will be an ipyton interpretor")
)

Configuration.add_application_properties(
    'autodoc', ['logging', 'doc', 'schema'],
    prog='AnyBlok auto documentation, version %r' % version,
)


def anyblok_createdb():
    """Create a database and install blok from config"""
    load_init_function_from_entry_points()
    Configuration.load('createdb')
    configuration_post_load()
    BlokManager.load()
    db_name = get_db_name()

    db_template_name = Configuration.get('db_template_name', None)
    url = Configuration.get('get_url')(db_name=db_name)
    create_database(url, template=db_template_name)
    registry = RegistryManager.get(db_name)
    if registry is None:
        return

    if Configuration.get('install_all_bloks'):
        bloks = registry.System.Blok.list_by_state('uninstalled')
    else:
        install_bloks = Configuration.get('install_bloks') or []
        iou_bloks = Configuration.get('install_or_update_bloks') or []
        bloks = list(set(install_bloks + iou_bloks))

    registry.upgrade(install=bloks)
    registry.commit()
    registry.close()


def anyblok_updatedb():
    """Update an existing database"""
    registry = anyblok.start('updatedb', loadwithoutmigration=True)

    installed_bloks = registry.System.Blok.list_by_state('installed')
    toupdate_bloks = registry.System.Blok.list_by_state('toupdate')
    required_install_bloks = []
    required_update_bloks = []
    for blok in (Configuration.get('install_or_update_bloks') or []):
        if blok in installed_bloks:
                required_update_bloks.append(blok)
        elif blok not in toupdate_bloks:
            required_install_bloks.append(blok)

    if Configuration.get('install_all_bloks'):
        install_bloks = registry.System.Blok.list_by_state('uninstalled')
    else:
        install_bloks = Configuration.get('install_bloks') or []
        install_bloks = list(set(install_bloks + required_install_bloks))

    if Configuration.get('update_all_bloks'):
        update_bloks = registry.System.Blok.list_by_state('installed')
    else:
        update_bloks = Configuration.get('update_bloks') or []
        update_bloks = list(set(update_bloks + required_update_bloks))

    uninstall_bloks = Configuration.get('uninstall_bloks')

    if registry:
        registry.update_blok_list()  # case, new blok added
        registry.upgrade(install=install_bloks, update=update_bloks,
                         uninstall=uninstall_bloks)
        registry.commit()
        registry.close()


def anyblok_nose():
    """Run nose unit test for the registry
    """
    registry = anyblok.start('nose', useseparator=True, unittest=True)

    defaultTest = []
    if registry:
        installed_bloks = registry.System.Blok.list_by_state("installed")
        selected_bloks = return_list(
            Configuration.get('selected_bloks')) or installed_bloks

        unwanted_bloks = return_list(
            Configuration.get('unwanted_bloks')) or []

        defaultTest = []
        for blok in installed_bloks:
            if blok not in selected_bloks or blok in unwanted_bloks:
                continue

            startpath = BlokManager.getPath(blok)
            for root, dirs, _ in walk(startpath):
                if 'tests' in dirs:
                    defaultTest.append(join(root, 'tests'))

        registry.close()  # free the registry to force create it again

    sys.exit(main(defaultTest=defaultTest))


def anyblok_interpreter():
    """Execute a script or open an interpreter
    """
    registry = anyblok.start('interpreter')
    if registry:
        registry.commit()
        python_script = Configuration.get('python_script')
        if python_script:
            with open(python_script, "r") as fh:
                exec(fh.read(), None, locals())
        else:
            try:
                from IPython import embed
                embed()
            except ImportError:
                import code
                code.interact(local=locals())


def anyblok2doc():
    """Return auto documentation for the registry
    """
    registry = anyblok.start('autodoc')
    if registry:
        registry.commit()
        doc = registry.Documentation()
        doc.auto_doc()
        if Configuration.get('doc_format') == 'RST':
            with open(Configuration.get('doc_output'), 'w') as fp:
                doc.toRST(fp)
        elif Configuration.get('doc_format') == 'UML':
            format_ = Configuration.get('schema_format')
            name_ = Configuration.get('schema_output')
            dot = ModelSchema(name_, format=format_)
            doc.toUML(dot)
            dot.save()
        elif Configuration.get('doc_format') == 'SQL':
            format_ = Configuration.get('schema_format')
            name_ = Configuration.get('schema_output')
            dot = SQLSchema(name_, format=format_)
            doc.toSQL(dot)
            dot.save()



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok import Declarations
from anyblok.column import Integer, String
from anyblok.relationship import One2Many


register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin


def _complete_one2many(**kwargs):
    primaryjoin = "ModelAddress.id == ModelPerson.address_id"

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))

    @register(Model)  # noqa
    class Address:

        persons = One2Many(model=Model.Person,
                           remote_columns="address_id",
                           primaryjoin=primaryjoin,
                           many2one="address")


def _minimum_one2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()
        persons = One2Many(model='Model.Person')

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))


def _minimum_one2many_remote_field_in_mixin(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Mixin)
    class MPerson:
        address_id = Integer(foreign_key=Model.Address.use('id'))

    @register(Model)
    class Person(Mixin.MPerson):

        name = String(primary_key=True)

    @register(Model)  # noqa
    class Address:

        persons = One2Many(model=Model.Person)


def _one2many_with_str_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))

    @register(Model)  # noqa
    class Address:

        persons = One2Many(model='Model.Person')


def _multi_fk_one2many():

    @register(Model)
    class Test:

        id = Integer(primary_key=True, unique=True)
        id2 = String(primary_key=True, unique=True)

    @register(Model)
    class Test2:

        id = Integer(primary_key=True)
        test_id = Integer(foreign_key=Model.Test.use('id'))
        test_id2 = String(foreign_key=Model.Test.use('id2'))

    @register(Model)  # noqa
    class Test:

        test2 = One2Many(model=Model.Test2, many2one="test")


class TestOne2Many(DBTestCase):

    def test_complete_one2many(self):
        registry = self.init_registry(_complete_one2many)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        address.persons.append(person)

        self.assertEqual(person.address, address)

    def test_complete_one2many_expire_field(self):
        registry = self.init_registry(_complete_one2many)
        self.assertIn(
            ('address',),
            registry.expire_attributes['Model.Person']['address_id'])
        self.assertIn(
            ('address', 'persons'),
            registry.expire_attributes['Model.Person']['address_id'])

    def test_minimum_one2many(self):
        registry = self.init_registry(_minimum_one2many)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        address.persons.append(person)

    def test_minimum_one2many_remote_field_in_mixin(self):
        registry = self.init_registry(_minimum_one2many)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        address.persons.append(person)

    def test_one2many_with_str_model(self):
        registry = self.init_registry(_one2many_with_str_model)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        address.persons.append(person)

    def test_same_model_backref(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                parent_id = Integer(foreign_key='Model.Test=>id')
                children = One2Many(model='Model.Test', many2one='parent')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        self.assertIs(t1.children[0], t2)
        self.assertIs(t2.parent, t1)

    def test_complet_with_multi_foreign_key(self):

        def add_in_registry():
            primaryjoin = "ModelTest.id == ModelTest2.test_id and "
            primaryjoin += "ModelTest.id2 == ModelTest2.test_id2"

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(foreign_key=Model.Test.use('id'))
                test_id2 = String(foreign_key=Model.Test.use('id2'))

            @register(Model)  # noqa
            class Test:

                test2 = One2Many(model=Model.Test2,
                                 remote_columns=['test_id', 'test_id2'],
                                 primaryjoin=primaryjoin,
                                 many2one="test")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test")
        t2 = registry.Test2.insert(test=t1)
        self.assertEqual(len(t1.test2), 1)
        self.assertIs(t1.test2[0], t2)

    def test_with_multi_foreign_key(self):
        registry = self.init_registry(_multi_fk_one2many)
        t1 = registry.Test.insert(id2="test")
        t2 = registry.Test2.insert(test=t1)
        self.assertEqual(len(t1.test2), 1)
        self.assertIs(t1.test2[0], t2)

    def test_with_multi_foreign_key_expire_field(self):
        registry = self.init_registry(_multi_fk_one2many)
        self.assertIn(
            ('test',),
            registry.expire_attributes['Model.Test2']['test_id'])
        self.assertIn(
            ('test', 'test2'),
            registry.expire_attributes['Model.Test2']['test_id'])
        self.assertIn(
            ('test',),
            registry.expire_attributes['Model.Test2']['test_id2'])
        self.assertIn(
            ('test', 'test2'),
            registry.expire_attributes['Model.Test2']['test_id2'])

    def test_with_multi_parallel_foreign_key_auto_detect(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address_1_id = Integer(foreign_key=Model.Address.use('id'))
                address_2_id = Integer(foreign_key=Model.Address.use('id'))

            @register(Model)  # noqa
            class Address:

                persons = One2Many(model=Model.Person)

        registry = self.init_registry(add_in_registry)
        address_1 = registry.Address.insert()
        address_2 = registry.Address.insert()
        person = registry.Person.insert(
            name='test', address_1_id=address_1.id, address_2_id=address_2.id)
        self.assertEqual(address_1.persons, [person])
        self.assertEqual(address_2.persons, [person])

    def test_with_multi_parallel_foreign_key_with_specific_primaryjoin(self):

        def add_in_registry():
            primaryjoin = (
                "or_("
                "ModelAddress.id == ModelPerson.address_1_id,"
                "ModelAddress.id == ModelPerson.address_2_id"
                ")"
            )

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address_1_id = Integer(foreign_key=Model.Address.use('id'))
                address_2_id = Integer(foreign_key=Model.Address.use('id'))

            @register(Model)  # noqa
            class Address:

                persons = One2Many(model=Model.Person,
                                   primaryjoin=primaryjoin)

        registry = self.init_registry(add_in_registry)
        address_1 = registry.Address.insert()
        address_2 = registry.Address.insert()
        person = registry.Person.insert(
            name='test', address_1_id=address_1.id, address_2_id=address_2.id)
        self.assertEqual(address_1.persons, [person])
        self.assertEqual(address_2.persons, [person])



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
"""Authorization subframework.

The founding principle of authorization handling within Anyblok is to check
authorization explicitely at the edge of the system (for instance,
for applications
exposed over HTTP, that would be in the controllers), because that is where the
idea of user, or slightly more generally, has functional semantics that can
be interpreted in the context of a given action.

In that spirit, we don't pass the user to the core framework and business
layers.
Instead, these provide *policies* to check permissions on records or query
records according to the.

The declarations at the edge will *associate* the policies with the
models. The edge user-aware methods will call the check and query facilities
provided by the core that themselves apply the relevant policies.
"""
from copy import deepcopy

from ..declarations import Declarations
from ..environment import EnvironmentManager
from ..registry import RegistryManager


@Declarations.add_declaration_type(isAnEntry=True,
                                   assemble='assemble_callback')
class AuthorizationBinding:
    """Encodes which policy to use per model or (model, permission).

    In the assembly phase, copies of the policy are issued, and the registry
    is set as an attribute on them. This is a bit memory inefficient, but
    otherwise, passing the registry would have to be in all AuthorizationRule
    API calls.
    """

    def __new__(cls, model_declaration, policy, permission=None):
        """Declare for given model that policy should be used.

        :param permission: if provided, the policy will apply for this
                           permission only, otherwise, it will act as the
                           default policy for this model.
        """
        cb = EnvironmentManager.get('current_blok')
        model = model_declaration.__registry_name__
        key = (model, permission) if permission is not None else model
        blok_declarations = RegistryManager.loaded_bloks[cb][cls.__name__]
        blok_declarations[key] = policy

    @classmethod
    def assemble_callback(cls, registry):
        policies = {}
        for blok in registry.ordered_loaded_bloks:
            policies.update(RegistryManager.loaded_bloks[blok][cls.__name__])

        # for this registry entry, the list of names is irrelevant pollution:
        del policies['registry_names']
        registry._authz_policies = deepcopy(policies)
        for policy in registry._authz_policies.values():
            policy.registry = registry



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from ..exceptions import CoreBaseException


class TestCoreBase(BlokTestCase):

    def test_to_primary_keys(self):
        with self.assertRaises(CoreBaseException):
            test = self.registry.System()
            test.to_primary_keys()

    def test_from_primary_keys(self):
        with self.assertRaises(CoreBaseException):
            self.registry.System.from_primary_keys()

    def test_get_primary_keys(self):
        with self.assertRaises(CoreBaseException):
            self.registry.System.get_primary_keys()

    def test_get_model(self):
        M = self.registry.System.Model
        M2 = self.registry.System.get_model('Model.System.Model')
        self.assertEqual(M, M2)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.declarations import Declarations, classmethod_cache
from anyblok.field import FieldException
from anyblok.column import Column
from anyblok.mapper import FakeColumn, FakeRelationShip
from anyblok.relationship import RelationShip, Many2Many
from anyblok.common import anyblok_column_prefix
from ..exceptions import SqlBaseException
from sqlalchemy.orm import aliased, ColumnProperty
from sqlalchemy.sql.expression import true
from sqlalchemy import or_, and_, inspect
from sqlalchemy_utils.models import NO_VALUE, NOT_LOADED_REPR
from sqlalchemy.orm.session import object_state


class uniquedict(dict):

    def add_in_res(self, key, attrs):
        if key not in self:
            self[key] = []

        for attr in attrs:
            if attr not in self[key]:
                self[key].append(attr)


class SqlMixin:

    def __repr__(self):
        state = inspect(self)
        field_reprs = []
        fields_description = self.fields_description()
        keys = list(fields_description.keys())
        keys.sort()
        for key in keys:
            type_ = fields_description[key]['type']
            if key in state.attrs:
                value = state.attrs.get(key).loaded_value
            elif (anyblok_column_prefix + key) in state.attrs:
                value = state.attrs.get(
                    anyblok_column_prefix + key).loaded_value
            else:
                continue

            if value == NO_VALUE:
                value = NOT_LOADED_REPR
            elif value and type_ in ('One2Many', 'Many2Many'):
                value = '<%s len(%d)>' % (
                    fields_description[key]['model'],
                    len(value)
                )
            elif value and type_ in ('One2One', 'Many2One'):
                value = '<%s(%s)>' % (
                    fields_description[key]['model'],
                    ', '.join(['='.join([x, str(y)])
                               for x, y in value.to_primary_keys().items()])
                )
            else:
                value = repr(value)

            field_reprs.append('='.join((key, value)))

        return '<%s(%s)>' % (
            self.__class__.__registry_name__, ', '.join(field_reprs)
        )

    @classmethod
    def define_table_args(cls):
        return ()

    @classmethod
    def define_mapper_args(cls):
        return {}

    @classmethod
    def get_all_registry_names(cls):
        models = [base.__registry_name__ for base in cls.__anyblok_bases__]
        models.insert(0, cls.__registry_name__)
        return models

    @classmethod
    def query(cls, *elements):
        """ Facility to do a SqlAlchemy query::

            query = MyModel.query()

        is equal at::

            query = self.registry.session.query(MyModel)

        :param elements: pass at the SqlAlchemy query, if the element is a
                         string then thet are see as field of the model
        :rtype: SqlAlchemy Query
        """
        res = []
        for f in elements:
            if isinstance(f, str):
                res.append(getattr(cls, f).label(f))
            else:
                res.append(f)

        if res:
            return cls.registry.query(*res)

        return cls.registry.query(cls)

    is_sql = True

    @classmethod
    def aliased(cls, *args, **kwargs):
        """ Facility to Apply an aliased on the model::

            MyModelAliased = MyModel.aliased()

        is equal at::

            from sqlalchemy.orm import aliased
            MyModelAliased = aliased(MyModel)

        :rtype: SqlAlchemy aliased of the model
        """
        return aliased(cls, *args, **kwargs)

    @classmethod
    def get_where_clause_from_primary_keys(cls, **pks):
        """ return the where clause to find object from pks

        :param \*\*pks: dict {primary_key: value, ...}
        :rtype: where clause
        :exception: SqlBaseException
        """
        _pks = cls.get_primary_keys()
        for pk in _pks:
            if pk not in pks:
                raise SqlBaseException("No primary key %s filled for %r" % (
                    pk, cls.__registry_name__))

        return [getattr(cls, k) == v for k, v in pks.items()]

    @classmethod
    def from_primary_keys(cls, **pks):
        """ return the instance of the model from the primary keys

        :param \*\*pks: dict {primary_key: value, ...}
        :rtype: instance of the model
        """
        where_clause = cls.get_where_clause_from_primary_keys(**pks)
        query = cls.query().filter(*where_clause)
        if query.count():
            return query.first()

        return None

    @classmethod
    def from_multi_primary_keys(cls, *pks):
        """ return the instances of the model from the primary keys

        :param \*pks: list of dict [{primary_key: value, ...}]
        :rtype: instances of the model
        """
        where_clause = []
        for _pks in pks:
            where_clause.append(cls.get_where_clause_from_primary_keys(**_pks))

        if not where_clause:
            return []

        where_clause = or_(*[and_(*x) for x in where_clause])

        query = cls.query().filter(where_clause)
        if query.count():
            return query.all()

        return []

    def to_primary_keys(self):
        """ return the primary keys and values for this instance

        :rtype: dict {primary key: value, ...}
        """
        pks = self.get_primary_keys()
        return {x: getattr(self, x) for x in pks}

    @classmethod_cache()
    def get_primary_keys(cls):
        """ return the name of the primary keys of the model

        :type: list of the primary keys name
        """
        C = cls.registry.System.Column
        query = C.query()
        query = query.filter(C.model.in_(cls.get_all_registry_names()))
        query = query.filter(C.primary_key == true())
        return query.all().name

    @classmethod_cache()
    def _fields_description(cls):
        """ Return the information of the Field, Column, RelationShip """
        Field = cls.registry.System.Field
        res = {}
        for base in cls.__anyblok_bases__:
            query = Field.query().filter(Field.model == base.__registry_name__)
            res.update({x.name: x._description() for x in query.all()})

        query = Field.query().filter(Field.model == cls.__registry_name__)
        res.update({x.name: x._description() for x in query.all()})
        return res

    @classmethod
    def fields_description(cls, fields=None):
        res = cls._fields_description()
        if fields:
            return {x: y for x, y in res.items() if x in fields}

        return res

    @classmethod_cache()
    def get_hybrid_property_columns(cls):
        """Return the hybrid properties columns name from the Model and the
        inherited model if they come from polymorphisme
        """
        hybrid_property_columns = cls.hybrid_property_columns
        if 'polymorphic_identity' in cls.__mapper_args__:
            pks = cls.get_primary_keys()
            fd = cls.fields_description(*pks)
            for pk in pks:
                if fd[pk].get('model'):
                    Model = cls.registry.get(fd[pk]['model'])
                    hybrid_property_columns.extend(
                        Model.get_hybrid_property_columns())

        return hybrid_property_columns

    def _format_field(self, field):
        related_fields = None
        if isinstance(field, (tuple, list)):
            if len(field) == 1:
                related_fields = ()
            elif len(field) == 2:
                related_fields = field[1]
                if related_fields is None:
                    related_fields = ()
                elif not isinstance(related_fields, (tuple, list)):
                    raise SqlBaseException("%r the related fields wanted "
                                           "must be a tuple or empty or "
                                           "None value" % related_fields)
            else:
                raise SqlBaseException("%r the number of argument is "
                                       "wrong, waiting 1 or 2 arguments "
                                       "(name of the relation[, (related "
                                       "fields)])" % (field,))

            field = field[0]

        return field, related_fields

    def to_dict(self, *fields):
        """ Transform a record to the dict of value

        :param fields: list of fields to put in dict; if not selected, fields
            then take them all. A field is either one of these:

                * a string (which is the name of the field)
                * a 2-tuple if the field is a relationship (name of the field,
                  tuple of foreign model fields)

        :rtype: dict

        Here are some examples::

            =>> instance.to_dict()  # get all fields
            {"id": 1,
             "column1": "value 1",
             "column2": "value 2",
             "column3": "value 3",
             "relation1": {"relation_pk_1": 42, "relation_pk_2": "also 42"}
                                            # m2o or o2o : this is a dictionary
             "relation2": [{"id": 28}, {"id": 1}, {"id": 34}]
                                # o2m or m2m : this is a list of dictionaries
             }

            =>> instance.to_dict("column1", "column2", "relation1")
                        # get selected fields only (without any constraints)
            {"column1": "value 1",
             "column2": "value 2",
             "relation1": {"relation_pk_1": 42, "relation_pk_2": "also 42"}
             }

            =>> instance.to_dict("column1", "column2", (
                        # select fields to use in the relation related model
                "relation1", ("relation_pk1", "name", "value")
                            # there is no constraints in the choice of fields
                ))
            {"column1": "value",
             "column2": "value",
             "relation1": {"relation_pk_1": 42, "name": "H2G2", "value": "42"}
             }

            =>> instance.to_dict("column1", "column2", ("relation1", ))
            # or
            =>> instance.to_dict("column1", "column2", ("relation1", None))
            # or
            =>> instance.to_dict("column1", "column2", ("relation1", ()))
                                  # select all the fields of the relation ship
            {"column1": "value",
             "column2": "value",
             "relation1": {"relation_pk_1": 42, "name": "H2G2", "value": "42"}
             }

            =>> instance.to_dict("column1", "column2", (
                                        # select relation fields recursively
                "relation1", ("name", "value", (
                    "relation", ("a", "b", "c")
                    ))
                ))
            {"column1": "value",
             "column2": "value",
             "relation1": {"name": "H2G2", "value": "42", "relation": [
                 {"a": 10, "b": 20, "c": 30},
                 {"a": 11, "b": 22, "c": 33},
                 ]}
             }
        """
        result = {}
        cls = self.__class__
        fields = fields if fields else cls.fields_description().keys()

        for field in fields:
            # if field is ("relation_name", ("list", "of", "relation",
            # "fields")), deal with it.
            field, related_fields = self._format_field(field)
            # Get the actual data
            field_value, field_property = getattr(self, field), None
            try:
                field_property = getattr(getattr(cls, field), 'property', None)
            except FieldException:
                pass

            # Deal with this data
            if field_property is None:
                # it is the case of field function (hyprid property)
                result[field] = field_value
            elif field_value is None or type(field_property) == ColumnProperty:
                # If value is None, then do not go any further whatever
                # the column property tells you.
                result[field] = field_value
            else:
                # it is should be RelationshipProperty
                if related_fields is None:
                    # If there is no field list to the relation,
                    # use only primary keys
                    related_fields = field_property.mapper.entity
                    related_fields = related_fields.get_primary_keys()
                # One2One, One2Many, Many2One or Many2Many ?
                if field_property.uselist:
                    result[field] = [r.to_dict(*related_fields)
                                     for r in field_value]
                else:
                    result[field] = field_value.to_dict(*related_fields)

        return result

    @classmethod_cache()
    def getFieldType(cls, name):
        """Return the type of the column

        ::

            TheModel.getFieldType(nameOfTheColumn)

        this method take care if it is a polymorphic model or not

        :param name: name of the column
        :rtype: String, the name of the Type of column used
        """
        Field = cls.registry.System.Field
        query = Field.query()
        query = query.filter(Field.name == name)
        query = query.filter(Field.model.in_(cls.get_all_registry_names()))
        query = query.limit(1)
        return query.one().ftype


def get_model_information(registry, registry_name):
    model = registry.loaded_namespaces_first_step[registry_name]
    for depend in model['__depends__']:
        if depend != registry_name:
            for x, y in get_model_information(registry, depend).items():
                if x not in model:
                    model[x] = y

    return model


@Declarations.register(Declarations.Core)
class SqlBase(SqlMixin):
    """ this class is inherited by all the SQL model
    """

    def get_modified_fields(self):
        """return the fields which have changed and their previous values"""
        state = object_state(self)
        modified_fields = {}
        for attr in state.manager.attributes:
            if not hasattr(attr.impl, 'get_history'):
                continue

            added, unmodified, deleted = attr.impl.get_history(
                state, state.dict)

            if added or deleted:
                field = attr.key
                if field.startswith(anyblok_column_prefix):
                    field = field[len(anyblok_column_prefix):]

                modified_fields[field] = deleted[0] if deleted else None

        return modified_fields

    def update(self, **values):
        """ Hight livel method to update the session for the instance
        ::

            self.update(val1=.., val2= ...)

        ..warning::

            the columns and values is passed as named arguments to show
            a difference with Query.update meth

        """
        for x, v in values.items():
            setattr(self, x, v)

        return 1 if values else 0

    @classmethod_cache()
    def find_remote_attribute_to_expire(cls, *fields):
        res = uniquedict()
        _fields = []
        _fields.extend(fields)
        model = get_model_information(cls.registry, cls.__registry_name__)
        while _fields:
            field = _fields.pop()
            field = field if isinstance(field, str) else field.name
            _field = model[field]

            if isinstance(_field, (Column, FakeColumn)):
                _fields.extend(x for x, y in model.items()
                               if (isinstance(y, RelationShip) and
                                   not isinstance(y, Many2Many))
                               for mapper in y.column_names
                               if mapper.attribute_name == field)
                if isinstance(_field, Column) and _field.foreign_key:
                    rmodel = cls.registry.loaded_namespaces_first_step[
                        _field.foreign_key.model_name]
                    for rc in [x for x, y in rmodel.items()
                               if isinstance(y, RelationShip)
                               for mapper in y.remote_columns
                               if mapper.attribute_name == field]:
                        rfield = rmodel[rc]
                        if isinstance(rfield, FakeRelationShip):
                            res.add_in_res(rfield.mapper.attribute_name, [rc])
                        elif (isinstance(rfield, RelationShip) and
                              'backref' in rfield.kwargs):
                            res.add_in_res(
                                rfield.kwargs['backref'][0], [rc])

            elif (isinstance(_field, RelationShip) and
                  not isinstance(_field, Many2Many) and
                  'backref' in _field.kwargs):
                res.add_in_res(field, [_field.kwargs['backref'][0]])
            elif isinstance(_field, FakeRelationShip):
                res.add_in_res(field, [_field.mapper.attribute_name])

        return res

    @classmethod_cache()
    def find_relationship(cls, *fields):
        """ Find column and relation ship link with the column or relationship
        passed in fields.

        :param \*fields: lists of the attribute name
        :rtype: list of the attribute name of the attribute and relation ship
        """
        res = []
        _fields = []
        _fields.extend(fields)
        model = get_model_information(cls.registry, cls.__registry_name__)
        while _fields:
            field = _fields.pop()
            if not isinstance(field, str):
                field = field.name

            if field in res:
                continue

            _field = model[field]
            res.append(field)
            if isinstance(_field, (Column, FakeColumn)):
                _fields.extend(x
                               for x, y in model.items()
                               if (isinstance(y, RelationShip) and
                                   not isinstance(y, Many2Many))
                               for mapper in y.column_names
                               if mapper.attribute_name == field)
            elif (isinstance(_field, RelationShip) and
                  not isinstance(_field, Many2Many)):
                for mapper in _field.column_names:
                    _fields.append(mapper.attribute_name)

        return res

    def expire_relationship_mapped(self, mappers):
        """ Expire the objects linked with this object, in function of
        the mappers definition
        """
        for field_name, rfields in mappers.items():
            fields = getattr(self, field_name)
            if not isinstance(fields, list):
                fields = [fields]

            for field in fields:
                if field is not None:
                    field.expire(*rfields)

    def refresh(self, *fields):
        """ Expire and reload all the attribute of the instance

        See: http://docs.sqlalchemy.org/en/latest/orm/session_api.html
        #sqlalchemy.orm.session.Session.refresh
        """
        self.registry.refresh(self, fields)

    def expunge(self):
        """Expunge the instance in the session"""
        self.registry.session.expunge(self)

    def expire(self, *fields):
        """ Expire the attribute of the instance, theses attributes will be
        load at the next  call of the instance

        see: http://docs.sqlalchemy.org/en/latest/orm/session_api.html
        #sqlalchemy.orm.session.Session.expire
        """
        self.registry.expire(self, fields)

    def delete(self, byquery=False, flush=True):
        """ Call the SqlAlchemy Query.delete method on the instance of the
        model::

            self.delete()

        is equal at::

            flush the session
            remove the instance of the session
            and expire all the session, to reload the relation ship

        """
        if byquery:
            cls = self.__class__
            cls.query().filter(*cls.get_where_clause_from_primary_keys(
                **self.to_primary_keys())).delete()
            self.expunge()
        else:
            model = self.registry.loaded_namespaces_first_step[
                self.__registry_name__]
            fields = model.keys()
            mappers = self.__class__.find_remote_attribute_to_expire(*fields)
            self.expire_relationship_mapped(mappers)
            self.registry.session.delete(self)

        if flush:
            self.registry.flush()

    @classmethod
    def insert(cls, **kwargs):
        """ Insert in the table of the model::

            MyModel.insert(...)

        is equal at::

            mymodel = MyModel(...)
            MyModel.registry.session.add(mymodel)
            MyModel.registry.flush()

        """
        instance = cls(**kwargs)
        cls.registry.add(instance)
        cls.registry.flush()
        return instance

    @classmethod
    def multi_insert(cls, *args):
        """ Insert in the table one or more entry of the model::

            MyModel.multi_insert([{...}, ...])

        the flush will be done only one time at the end of the insert

        :exception: SqlBaseException
        """
        instances = cls.registry.InstrumentedList()
        for kwargs in args:
            if not isinstance(kwargs, dict):
                raise SqlBaseException("multi_insert method wait list of dict")

            instance = cls(**kwargs)
            cls.registry.add(instance)
            instances.append(instance)

        if instances:
            cls.registry.flush()

        return instances



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations, reload_module_if_blok_is_reloading


@Declarations.register(Declarations.Mixin)
class DocElement:

    def _auto_doc(self, Model, elements, *args, **kwargs):
        for el in Model.getelements(*args, **kwargs):
            _el = Model(el)
            if _el.exist(*args, **kwargs):
                elements.append(_el)

    def _toRST(self, doc, Model, elements):
        Model.header2RST(doc)
        for el in elements:
            el.toRST(doc)

        Model.footer2RST(doc)


@Declarations.register(Declarations.Model)
class Documentation(Declarations.Mixin.DocElement):

    def __init__(self):
        self.bloks = []
        self.models = []

    def auto_doc_blok(self):
        self._auto_doc(self.registry.Documentation.Blok, self.bloks)

    def auto_doc_model(self):
        self._auto_doc(self.registry.Documentation.Model, self.models)

    def auto_doc(self):
        self.auto_doc_blok()
        self.auto_doc_model()

    def header2RST(self, doc):
        pass

    def footer2RST(self, doc):
        pass

    def chapter2RST(self, doc):
        self.toRST_blok(doc)
        self.toRST_model(doc)

    def toRST(self, doc):
        title = 'Documentation of the %s project' % self.Env.get('db_name')
        quote = "=" * len(title)
        doc.write('\n'.join([quote, title, quote, '\n']))
        self.header2RST(doc)
        self.chapter2RST(doc)
        self.footer2RST(doc)

    def toUML(self, dot):
        for m in self.models:
            m.toUML_add_model(dot)

        for m in self.models:
            m.toUML_add_attributes(dot)

    def toSQL(self, dot):
        for m in self.models:
            m.toSQL_add_table(dot)

        for m in self.models:
            m.toSQL_add_fields(dot)

    def toRST_blok(self, doc):
        self._toRST(doc, self.registry.Documentation.Blok, self.bloks)

    def toRST_model(self, doc):
        self._toRST(doc, self.registry.Documentation.Model, self.models)


from . import blok  # noqa
reload_module_if_blok_is_reloading(blok)
from . import model  # noqa
reload_module_if_blok_is_reloading(model)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .testcase import DBTestCase
from ..authorization.rule.base import deny_all
from ..authorization.rule.base import RuleNotForModelClasses
from ..authorization.rule.attraccess import AttributeAccessRule
from anyblok.test_bloks.authorization import TestRuleOne
from anyblok.test_bloks.authorization import TestRuleTwo


class TestAuthorizationDeclaration(DBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_association(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok7',))
        record = registry.Test(id=23, label='Hop')
        self.assertIsInstance(registry.lookup_policy(record, 'Read'),
                              TestRuleOne)
        self.assertIsInstance(registry.lookup_policy(record, 'Other'),
                              TestRuleTwo)

        record = registry.Test2(id=2, label='Hop')
        self.assertIs(registry.lookup_policy(record, 'Read'), deny_all)

    def test_override(self):
        # test-blok8 depends on test-blok7
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok8',))
        # lookup can be made on model itself
        model = registry.Test
        self.assertIsInstance(registry.lookup_policy(model, 'Read'),
                              TestRuleOne)
        self.assertIsInstance(registry.lookup_policy(model, 'Other'),
                              TestRuleOne)
        self.assertIsInstance(registry.lookup_policy(model, 'Write'),
                              TestRuleTwo)

    def test_model_based_policy(self):
        """Test the model based policy using the default Grant model.

        The policy is defined in the ``model_authz`` blok
        The supporting default model is installed by the same blok.
        #TODO move this test to the 'model_authz' blok
        """
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok9',))
        model = registry.Test2
        Grant = registry.Authorization.ModelPermissionGrant
        Grant.insert(model='Model.Test2',
                     principal="Franck",
                     permission="Read")

        record = model.insert(id=2)
        self.assertTrue(
            registry.check_permission(record, ('Franck',), 'Read'))
        self.assertFalse(
            registry.check_permission(record, ('Franck',), 'Write'))

        # With this policy, permission can be checked on the model
        self.assertTrue(
            registry.check_permission(model, ('Franck',), 'Read'))
        self.assertFalse(
            registry.check_permission(model, ('Franck',), 'Write'))

        # This can be also called directly from the model class
        self.assertTrue(model.has_model_perm(('Franck',), 'Read'))
        self.assertFalse(model.has_model_perm(('Franck',), 'Write'))
        with self.assertRaises(TypeError):
            model.has_perm(('Franck',), 'Write')

        query = model.query().filter(model.id != 1)
        self.assertEqual(query.count(), 1)

        filtered = registry.wrap_query_permission(
            query, ('Franck',), 'Read')
        self.assertEqual(filtered.count(), 1)
        self.assertEqual(filtered.first().id, 2)
        all_results = filtered.all()
        self.assertEqual(all_results[0].id, 2)

        filtered = registry.wrap_query_permission(
            query, ('Franck',), 'Write')
        self.assertEqual(filtered.count(), 0)
        self.assertIsNone(filtered.first())
        self.assertEqual(len(filtered.all()), 0)

        # the same can be achieved from the query
        filtered = query.with_perm(('Franck',), 'Read')
        self.assertEqual(filtered.count(), 1)
        self.assertEqual(filtered.first().id, 2)
        all_results = filtered.all()
        self.assertEqual(all_results[0].id, 2)

        filtered = query.with_perm(('Franck',), 'Write')
        self.assertEqual(filtered.count(), 0)
        self.assertIsNone(filtered.first())
        self.assertEqual(len(filtered.all()), 0)

    def test_attr_based_policy(self):
        """Test the attribute based policy, in conjunction with the model one.

        The policy is defined in the ``attr_authz`` blok
        The supporting default model is installed by the same blok.
        TODO move this test to the 'attr_authz' blok, or put the policy in the
        main code body.
        """
        registry = self.init_registry(None)
        registry.upgrade(install=('test-blok10',))
        model = registry.Test2
        Grant = registry.Authorization.ModelPermissionGrant
        Grant.insert(model='Model.Test2',
                     principal="Franck",
                     permission="Read")
        Grant.insert(model='Model.Test2',
                     principal="Georges",
                     permission="Read")

        # We also test that this entry is ignored, because it is
        # superseded by attribute-based policy
        Grant.insert(model='Model.Test',
                     principal="Franck",
                     permission="Write")

        record = model.insert(id=1, owner='Georges')
        self.assertTrue(
            registry.check_permission(record, ('Franck',), 'Read'))
        self.assertTrue(
            registry.check_permission(record, ('Georges',), 'Write'))
        self.assertFalse(
            registry.check_permission(record, ('Franck',), 'Write'))

        # The same checks can be done from the record
        self.assertTrue(record.has_perm(('Franck',), 'Read'))
        self.assertTrue(record.has_perm(('Georges',), 'Write'))
        self.assertFalse(record.has_perm(('Franck',), 'Write'))

        # With this policy, permission cannot be checked on the model
        with self.assertRaises(RuleNotForModelClasses) as arc:
            registry.check_permission(model, ('Franck',), 'Write')
        self.assertIsInstance(arc.exception.policy,
                              AttributeAccessRule)

        # ... unless one defines a model_rule to handle the case
        self.assertFalse(registry.check_permission(
            model, ('Franck',), 'PermWithModelRule'))

        Grant = registry.Authorization.ModelPermissionGrant
        Grant.insert(model=model.__registry_name__,
                     principal="Franck",
                     permission="PermWithModelRule")
        self.assertTrue(registry.check_permission(
            model, ('Franck',), 'PermWithModelRule'))
        self.assertFalse(record.has_perm(('Franck',), 'PermWithModelRule'))

        # Query wrapping tests

        model.insert(id=2, owner='Franck')
        model.insert(id=3, owner='JS')

        query = model.query()
        self.assertEqual(query.count(), 3)

        filtered = registry.wrap_query_permission(
            query, ('Franck',), 'Write')

        self.assertEqual(filtered.count(), 1)
        self.assertEqual(filtered.first().id, 2)
        all_results = filtered.all()
        self.assertEqual(all_results[0].id, 2)

        # The same can be achieved directly on query
        filtered = query.with_perm(('Franck',), 'Write')
        self.assertEqual(filtered.count(), 1)
        self.assertEqual(filtered.first().id, 2)
        all_results = filtered.all()
        self.assertEqual(all_results[0].id, 2)

        filtered = registry.wrap_query_permission(
            query, ('Franck',), 'Read')
        self.assertEqual(filtered.count(), 3)

        filtered = registry.wrap_query_permission(
            query, ('Franck', 'Georges',), 'Write')
        self.assertEqual(filtered.count(), 2)
        self.assertEqual([r.id for r in filtered.all()], [1, 2])



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import String, Boolean


register = Declarations.register
System = Declarations.Model.System
Mixin = Declarations.Mixin


@register(System)
class Column(System.Field):

    name = String(primary_key=True)
    model = String(primary_key=True)
    autoincrement = Boolean(label="Auto increment")
    foreign_key = String()
    primary_key = Boolean()
    unique = Boolean()
    nullable = Boolean()
    remote_model = String()

    def _description(self):
        res = super(Column, self)._description()
        res.update(nullable=self.nullable, primary_key=self.primary_key,
                   model=self.remote_model)
        return res

    @classmethod
    def get_cname(self, field, cname):
        """ Return the real name of the column

        :param field: the instance of the column
        :param cname: Not use here
        :rtype: string of the real column name
        """
        return cname

    @classmethod
    def add_field(cls, cname, column, model, table, ftype):
        """ Insert a column definition

        :param cname: name of the column
        :param column: instance of the column
        :param model: namespace of the model
        :param table: name of the table of the model
        :param ftype: type of the AnyBlok Field
        """
        c = column.property.columns[0]
        autoincrement = c.autoincrement
        if autoincrement == 'auto':
            autoincrement = (True
                             if c.primary_key and ftype == 'Integer'
                             else False)

        vals = dict(autoincrement=autoincrement,
                    code=table + '.' + cname,
                    model=model, name=cname,
                    foreign_key=c.info.get('foreign_key'),
                    label=c.info.get('label'),
                    nullable=c.nullable,
                    primary_key=c.primary_key,
                    ftype=ftype,
                    remote_model=c.info.get('remote_model'),
                    unique=c.unique)
        cls.insert(**vals)

    @classmethod
    def alter_field(cls, column, meta_column, ftype):
        """ Update an existing column

        :param column: instance of the Column model to update
        :param meta_column: instance of the SqlAlchemy column
        :param ftype: type of the AnyBlok Field
        """
        c = meta_column.property.columns[0]
        autoincrement = c.autoincrement
        if autoincrement == 'auto':
            autoincrement = (True
                             if c.primary_key and ftype == 'Integer'
                             else False)

        if column.autoincrement != autoincrement:
            column.autoincrement = autoincrement

        for col in ('nullable', 'primary_key', 'unique'):
            if getattr(column, col) != getattr(c, col):
                setattr(column, col, getattr(c, col))

        for col in ('foreign_key', 'label', 'remote_model'):
            if getattr(column, col) != c.info.get(col):
                setattr(column, col, c.info.get(col))

        if column.ftype != ftype:
            column.ftype = ftype



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.declarations import Declarations, cache
from anyblok.column import Integer, String


register = Declarations.register
Model = Declarations.Model


@register(Model)
class Test:

    id = Integer(primary_key=True)
    blok = String()
    mode = String()

    @cache()
    def test_with_cache(self):
        pass


@register(Model.System)
class Blok:

    def install(self):
        super(Blok, self).install()
        self.registry.Test.insert(blok=self.name, mode='install')

    def upgrade(self):
        super(Blok, self).upgrade()
        self.registry.Test.insert(blok=self.name, mode='update')

    def load(self):
        super(Blok, self).load()
        self.registry.Test.insert(blok=self.name, mode='load')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import logging
from functools import wraps
from anyblok.environment import EnvironmentManager
from .common import function_name


BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, _N, DEFAULT = range(10)
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
COLOR_PATTERN = "%s%s%%s%s" % (COLOR_SEQ, COLOR_SEQ, RESET_SEQ)
LEVEL_COLOR_MAPPING = {
    logging.DEBUG: (BLUE, DEFAULT),
    logging.INFO: (GREEN, DEFAULT),
    logging.WARNING: (YELLOW, DEFAULT),
    logging.ERROR: (RED, DEFAULT),
    logging.CRITICAL: (WHITE, RED),
}


class consoleFormatter(logging.Formatter):
    """ Define the format for console logging """

    def format(self, record):
        """ Add color to the message

        :param record: logging record instance
        :rtype: logging record formatted
        """
        fg_color, bg_color = LEVEL_COLOR_MAPPING[record.levelno]
        record.levelname = COLOR_PATTERN % (
            30 + fg_color, 40 + bg_color, record.levelname)
        fg_color, bg_color = CYAN, DEFAULT
        record.database = COLOR_PATTERN % (
            30 + fg_color, 40 + bg_color,
            EnvironmentManager.get('db_name', 'No database'))

        return logging.Formatter.format(self, record)


class anyblokFormatter(logging.Formatter):
    """ Define the format for console logging """

    def format(self, record):
        """ Add color to the message

        :param record: logging record instance
        :rtype: logging record formatted
        """
        record.database = EnvironmentManager.get('db_name', 'No database')
        return logging.Formatter.format(self, record)


def log(logger, level='info', withargs=False):
    """ decorator to log the entry of a method

    There are 5 levels of logging
    * debug
    * info (default)
    * warning
    * error
    * critical

    example::

        from logging import getLogger
        logger = getLogger(__name__)

        @log(logger)
        def foo(...):
            ...

    :param level: AnyBlok log level
    :param withargs: If True, add args and kwargs in the log message
    """

    def wrapper(function):
        @wraps(function)
        def f(*args, **kwargs):
            if level == 'debug' or withargs:
                getattr(logger, level)("%s with args %r and kwargs %r" % (
                    function_name(function), args, kwargs))
            else:
                getattr(logger, level)(function_name(function))

            return function(*args, **kwargs)

        return f

    return wrapper



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.


class CoreBaseException(TypeError):
    """ Exception for ``Core.Base`` """


class SqlBaseException(Exception):
    """ Simple Exception for sql base """


class QueryException(Exception):
    """ Simple Exception for query """


class CacheException(Exception):
    """ Simple Exception for the cache Model """


class ParameterException(Exception):
    """ Simple exception for System.Parameter """


class ForbidUpdateException(Exception):
    """ Simple exception for Mixin.ForbiddenUpdate """


class ForbidDeleteException(Exception):
    """ Simple exception for Mixin.ForbiddenDelete """



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase, DBTestCase
from anyblok.registry import RegistryManager
from anyblok import Declarations
from anyblok.environment import EnvironmentManager
from anyblok.column import Integer, String, Boolean
from anyblok.bloks.anyblok_core.exceptions import (
    ForbidDeleteException, ForbidUpdateException
)
register = Declarations.register
unregister = Declarations.unregister
Mixin = Declarations.Mixin
Model = Declarations.Model


class OneInterface:
    pass


class TestCoreInterfaceMixin(TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestCoreInterfaceMixin, cls).setUpClass()
        RegistryManager.init_blok('testMixin')
        EnvironmentManager.set('current_blok', 'testMixin')

    @classmethod
    def tearDownClass(cls):
        super(TestCoreInterfaceMixin, cls).tearDownClass()
        EnvironmentManager.set('current_blok', None)
        del RegistryManager.loaded_bloks['testMixin']

    def setUp(self):
        super(TestCoreInterfaceMixin, self).setUp()
        blokname = 'testMixin'
        RegistryManager.loaded_bloks[blokname]['Mixin'] = {
            'registry_names': []}

    def assertInMixin(self, *args):
        blokname = 'testMixin'
        blok = RegistryManager.loaded_bloks[blokname]
        self.assertEqual(len(blok['Mixin']['Mixin.MyMixin']['bases']),
                         len(args))
        for cls_ in args:
            has = cls_ in blok['Mixin']['Mixin.MyMixin']['bases']
            self.assertEqual(has, True)

    def assertInRemoved(self, cls):
        core = RegistryManager.loaded_bloks['testMixin']['removed']
        if cls in core:
            return True

        self.fail('Not in removed')

    def test_add_interface(self):
        register(Mixin, cls_=OneInterface, name_='MyMixin')
        self.assertEqual('Mixin', Mixin.MyMixin.__declaration_type__)
        self.assertInMixin(OneInterface)
        dir(Declarations.Mixin.MyMixin)

    def test_add_interface_with_decorator(self):

        @register(Mixin)
        class MyMixin:
            pass

        self.assertEqual('Mixin', Mixin.MyMixin.__declaration_type__)
        self.assertInMixin(MyMixin)

    def test_add_two_interface(self):

        register(Mixin, cls_=OneInterface, name_="MyMixin")

        @register(Mixin)
        class MyMixin:
            pass

        self.assertInMixin(OneInterface, MyMixin)

    def test_remove_interface_with_1_cls_in_registry(self):

        register(Mixin, cls_=OneInterface, name_="MyMixin")
        self.assertInMixin(OneInterface)
        unregister(Mixin.MyMixin, OneInterface)

        blokname = 'testMixin'
        self.assertEqual(hasattr(Mixin, blokname), False)
        self.assertInMixin(OneInterface)
        self.assertInRemoved(OneInterface)

    def test_remove_interface_with_2_cls_in_registry(self):

        register(Mixin, cls_=OneInterface, name_="MyMixin")

        @register(Mixin)
        class MyMixin:
            pass

        self.assertInMixin(OneInterface, MyMixin)
        unregister(Mixin.MyMixin, OneInterface)
        self.assertInMixin(MyMixin, OneInterface)
        self.assertInRemoved(OneInterface)


class TestGivenMixin(DBTestCase):

    def test_forbidden_delete(self):

        def add_in_registry():

            @register(Model)
            class Test(Mixin.ForbidDelete):

                id = Integer(primary_key=True)

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        with self.assertRaises(ForbidDeleteException):
            t.delete()

    def test_forbidden_update(self):

        def add_in_registry():

            @register(Model)
            class Test(Mixin.ForbidUpdate):

                id = Integer(primary_key=True)
                name = String()

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        t.name = 'test'
        with self.assertRaises(ForbidUpdateException):
            registry.flush()

    def test_readonly_delete(self):

        def add_in_registry():

            @register(Model)
            class Test(Mixin.ReadOnly):

                id = Integer(primary_key=True)

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        with self.assertRaises(ForbidDeleteException):
            t.delete()

    def test_readonly_update(self):

        def add_in_registry():

            @register(Model)
            class Test(Mixin.ReadOnly):

                id = Integer(primary_key=True)
                name = String()

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        t.name = 'test'
        with self.assertRaises(ForbidUpdateException):
            registry.flush()

    def test_conditional_forbidden_delete(self):

        def add_in_registry():

            @register(Model)
            class Test(Mixin.ConditionalForbidDelete):

                id = Integer(primary_key=True)
                forbid_delete = Boolean(default=False)

                def check_if_forbid_delete_condition_is_true(self):
                    return self.forbid_delete

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(forbid_delete=True)
        t1.delete()

        with self.assertRaises(ForbidDeleteException):
            t2.delete()

    def test_conditional_forbidden_delete_mission_method(self):

        def add_in_registry():

            @register(Model)
            class Test(Mixin.ConditionalForbidDelete):

                id = Integer(primary_key=True)

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        with self.assertRaises(ForbidDeleteException):
            t.delete()

    def test_conditional_forbidden_update_missing_method(self):

        def add_in_registry():

            @register(Model)
            class Test(Mixin.ConditionalForbidUpdate):

                id = Integer(primary_key=True)
                name = String()

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        t.name = 'test'
        with self.assertRaises(ForbidUpdateException):
            registry.flush()

    def test_conditional_forbidden_update(self):

        def add_in_registry():

            @register(Model)
            class Test(Mixin.ConditionalForbidUpdate):

                id = Integer(primary_key=True)
                forbid_update = Boolean(default=False)
                name = String()

                def check_if_forbid_update_condition_is_true(self, **changed):
                    return self.forbid_update

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t1.name = 'test'
        registry.flush()
        t2 = registry.Test.insert(forbid_update=True)
        t2.name = 'test'
        with self.assertRaises(ForbidUpdateException):
            registry.flush()

    def add_in_registry_conditional_readonly(self):

        @register(Model)
        class Test(Mixin.ConditionalReadOnly):

            id = Integer(primary_key=True)
            readonly = Boolean(default=False)
            name = String()

            def check_if_forbid_update_condition_is_true(self,
                                                         **previous_values):
                return previous_values.get('readonly', self.readonly)

            def check_if_forbid_delete_condition_is_true(self):
                return self.readonly

    def test_conditional_readonly_delete(self):
        registry = self.init_registry(self.add_in_registry_conditional_readonly)
        t1 = registry.Test.insert()
        t1.delete()
        t2 = registry.Test.insert(readonly=True)
        with self.assertRaises(ForbidDeleteException):
            t2.delete()

    def test_conditional_readonly_update(self):
        registry = self.init_registry(self.add_in_registry_conditional_readonly)
        t1 = registry.Test.insert()
        t1.name = 'test'
        registry.flush()
        t2 = registry.Test.insert(readonly=True)
        t2.name = 'test'
        with self.assertRaises(ForbidUpdateException):
            registry.flush()

    def test_boolean_forbidden_delete(self):

        def add_in_registry():

            @register(Model)
            class Test(Mixin.BooleanForbidDelete):

                id = Integer(primary_key=True)

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(forbid_delete=True)
        t1.delete()

        with self.assertRaises(ForbidDeleteException):
            t2.delete()

    def test_boolean_forbidden_update(self):

        def add_in_registry():

            @register(Model)
            class Test(Mixin.BooleanForbidUpdate):

                id = Integer(primary_key=True)
                name = String()

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t1.name = 'test'
        registry.flush()
        t2 = registry.Test.insert(forbid_update=True)
        t2.name = 'test'
        with self.assertRaises(ForbidUpdateException):
            registry.flush()

    def add_in_registry_boolean_readonly(self):

        @register(Model)
        class Test(Mixin.BooleanReadOnly):

            id = Integer(primary_key=True)
            name = String()

    def test_boolean_readonly_delete(self):
        registry = self.init_registry(self.add_in_registry_boolean_readonly)
        t1 = registry.Test.insert()
        t1.delete()
        t2 = registry.Test.insert(readonly=True)
        with self.assertRaises(ForbidDeleteException):
            t2.delete()

    def test_boolean_readonly_update(self):
        registry = self.init_registry(self.add_in_registry_boolean_readonly)
        t1 = registry.Test.insert()
        t1.name = 'test'
        registry.flush()
        t2 = registry.Test.insert(readonly=True)
        t2.name = 'test'
        with self.assertRaises(ForbidUpdateException):
            registry.flush()

    def add_in_registry_state_readonly(self):

        @register(Model)
        class Test(Mixin.StateReadOnly):
            DEFAULT_STATE = 'draft'

            @classmethod
            def get_states(cls):
                return {
                    'draft': 'Draft',
                    'started': 'Started',
                    'done': 'Done',
                }

            def check_if_forbid_update_condition_is_true(self, **changed):
                if 'state' in changed:
                    return False

                return self.state == 'done'

            def check_if_forbid_delete_condition_is_true(self):
                return self.state != 'draft'

            id = Integer(primary_key=True)
            name = String()

    def test_state_readonly_delete_1(self):
        registry = self.init_registry(self.add_in_registry_state_readonly)
        t = registry.Test.insert()
        t.delete()

    def test_state_readonly_delete_2(self):
        registry = self.init_registry(self.add_in_registry_state_readonly)
        t = registry.Test.insert(state='started')
        with self.assertRaises(ForbidDeleteException):
            t.delete()

    def test_state_readonly_delete_3(self):
        registry = self.init_registry(self.add_in_registry_state_readonly)
        t = registry.Test.insert(state='done')
        with self.assertRaises(ForbidDeleteException):
            t.delete()

    def test_state_readonly_update_1(self):
        registry = self.init_registry(self.add_in_registry_state_readonly)
        t1 = registry.Test.insert()
        t1.name = 'test'
        registry.flush()

    def test_state_readonly_update_2(self):
        registry = self.init_registry(self.add_in_registry_state_readonly)
        t = registry.Test.insert(state='done')
        t.name = 'test'
        with self.assertRaises(ForbidUpdateException):
            registry.flush()

    def test_state_readonly_update_3(self):
        registry = self.init_registry(self.add_in_registry_state_readonly)
        t = registry.Test.insert()
        t.name = 'test1'
        t.state = 'started'
        registry.flush()
        t.name = 'test2'
        t.state = 'done'
        registry.flush()
        t.name = 'test3'
        with self.assertRaises(ForbidUpdateException):
            registry.flush()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase


class TestSystemField(BlokTestCase):

    pass  # No test for the moment



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.relationship import Many2One
from anyblok.column import String, Selection
from io import StringIO
from csv import DictWriter
from .exceptions import CSVExporterException


register = Declarations.register
IO = Declarations.Model.IO
Mixin = Declarations.Mixin


@register(IO)
class Exporter(Mixin.IOCSVMixin):

    @classmethod
    def get_mode_choices(cls):
        res = super(Exporter, cls).get_mode_choices()
        res.update({'Model.IO.Exporter.CSV': 'CSV'})
        return res


@register(IO.Exporter)
class Field(Mixin.IOCSVFieldMixin):

    exporter = Many2One(model=IO.Exporter, nullable=False,
                        one2many='fields_to_export')
    mode = Selection(selections='get_selection', nullable=False, default='any')
    mapping = String()

    @classmethod
    def get_selection(cls):
        return {
            'any': '',
            'external_id': 'External ID',
        }

    def _get_fields_description(self, name, entry):
        Model = self.get_model(entry.__registry_name__)
        fields_description = Model.fields_description(fields=[name])
        if name not in fields_description:
            raise CSVExporterException(
                "unknow field %r in exporter field %r" % (
                    name, self.name))

        return fields_description[name]

    def _value2str(self, exporter, name, entry, external_id):
        fields_description = self._get_fields_description(name, entry)
        if fields_description['primary_key'] and external_id:
            return self.registry.IO.Exporter.get_key_mapping(entry)

        ctype = fields_description['type']
        model = fields_description['model']
        return exporter.value2str(getattr(entry, name), ctype,
                                  external_id=external_id, model=model)

    def _rc_get_sub_entry(self, name, entry):
        fields_description = self._get_fields_description(name, entry)
        if fields_description['type'] in ('Many2One', 'One2One'):
            return getattr(entry, name)

        elif fields_description['model']:
            model = fields_description['model']
            Model = self.registry.get(model)
            pks = Model.get_primary_keys()
            if len(pks) == 1:
                pks = {pks[0]: getattr(entry, name)}
            else:
                raise CSVExporterException(
                    "Not implemented yet")

            return Model.from_primary_keys(**pks)

        else:
            raise CSVExporterException(
                "the field %r of %r is not in (Many2One, One2One) "
                "or has not a foreign key")

    def value2str(self, exporter, entry):

        def _rc_get_value(names, entry):
            if not names:
                return ''
            elif len(names) == 1:
                external_id = False if self.mode == 'any' else True
                return self._value2str(exporter, names[0], entry, external_id)
            else:
                return _rc_get_value(
                    names[1:], self._rc_get_sub_entry(names[0], entry))

        return _rc_get_value(self.name.split('.'), entry)

    def format_header(self):
        if self.mode == 'any':
            return self.name
        else:
            return self.name + '/EXTERNAL_ID'


@register(IO.Exporter)
class CSV:

    def __init__(self, exporter):
        self.exporter = exporter

    @classmethod
    def insert(cls, delimiter=None, quotechar=None, fields=None, **kwargs):
        kwargs['mode'] = cls.__registry_name__
        if 'model' in kwargs:
            if not isinstance(kwargs['model'], str):
                kwargs['model'] = kwargs['model'].__registry_name__

        if delimiter is not None:
            kwargs['csv_delimiter'] = delimiter

        if quotechar is not None:
            kwargs['csv_quotechar'] = quotechar

        exporter = cls.registry.IO.Exporter.insert(**kwargs)
        if fields:
            for field in fields:
                field['exporter'] = exporter

            cls.registry.IO.Exporter.Field.multi_insert(*fields)

        return exporter

    def get_header(self):
        return [field.format_header() for
                field in self.exporter.fields_to_export]

    def run(self, entries):
        csvfile = StringIO()
        writer = DictWriter(csvfile, fieldnames=self.get_header(),
                            delimiter=self.exporter.csv_delimiter,
                            quotechar=self.exporter.csv_quotechar)
        writer.writeheader()
        for entry in entries:
            writer.writerow(
                {field.format_header(): field.value2str(self.exporter, entry)
                 for field in self.exporter.fields_to_export})

        csvfile.seek(0)
        return csvfile



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from ..exceptions import IOMappingSetException


class TestIOMapping(BlokTestCase):

    def setUp(self):
        super(TestIOMapping, self).setUp()
        self.Mapping = self.registry.IO.Mapping
        self.Model = self.registry.System.Model
        self.Column = self.registry.System.Column
        self.Field = self.registry.System.Field
        self.Blok = self.registry.System.Blok

    def test_set_primary_key(self):
        model = self.Model.query().first()
        res = self.Mapping.set_primary_keys(
            model.__registry_name__, 'test_set_pk', dict(name=model.name))
        query = self.Mapping.query()
        query = query.filter(self.Mapping.key == 'test_set_pk')
        self.assertEqual(query.count(), 1)
        mapping = query.first()
        self.assertEqual(mapping, res)
        self.assertEqual(mapping.model, model.__registry_name__)
        self.assertEqual(mapping.primary_key, dict(name=model.name))

    def test_set_primary_keys(self):
        column = self.Column.query().first()
        res = self.Mapping.set_primary_keys(
            column.__registry_name__, 'test_set_pks',
            dict(model=column.model, name=column.name))
        query = self.Mapping.query()
        query = query.filter(self.Mapping.key == 'test_set_pks')
        self.assertEqual(query.count(), 1)
        mapping = query.first()
        self.assertEqual(mapping, res)
        self.assertEqual(mapping.model, column.__registry_name__)
        self.assertEqual(mapping.primary_key,
                         dict(model=column.model, name=column.name))

    def test_set(self):
        column = self.Column.query().first()
        res = self.Mapping.set('test_set', column)
        query = self.Mapping.query()
        query = query.filter(self.Mapping.key == 'test_set')
        self.assertEqual(query.count(), 1)
        mapping = query.first()
        self.assertEqual(mapping, res)
        self.assertEqual(mapping.model, column.__registry_name__)
        self.assertEqual(mapping.primary_key,
                         dict(model=column.model, name=column.name))

    def test_get_primary_key(self):
        model = self.Model.query().first()
        self.Mapping.set_primary_keys(
            model.__registry_name__, 'test_get_pk', dict(name=model.name))
        mapping = self.Mapping.get_mapping_primary_keys(
            model.__registry_name__, 'test_get_pk')
        self.assertEqual(mapping, dict(name=model.name))

    def test_get_primary_keys(self):
        column = self.Column.query().first()
        self.Mapping.set_primary_keys(
            column.__registry_name__, 'test_get_pks',
            dict(model=column.model, name=column.name))
        mapping = self.Mapping.get_mapping_primary_keys(
            column.__registry_name__, 'test_get_pks')
        self.assertEqual(mapping, dict(model=column.model, name=column.name))

    def test_get(self):
        column = self.Column.query().first()
        self.Mapping.set('test_get', column)
        mapping = self.Mapping.get(column.__registry_name__, 'test_get')
        self.assertEqual(mapping, column)

    def test_delete(self):
        column = self.Column.query().first()
        self.Mapping.set('test_delete', column)
        mapping = self.Mapping.get(column.__registry_name__, 'test_delete')
        self.assertEqual(mapping, column)
        self.Mapping.delete(column.__registry_name__, 'test_delete')
        mapping = self.Mapping.get(column.__registry_name__, 'test_delete')
        self.assertEqual(mapping, None)

    def test_multi_delete(self):
        columns = {'test_%s' % m.code: m
                   for m in self.Column.query().limit(5).all()}
        model = self.Column.__registry_name__

        # create all
        for key, instance in columns.items():
            self.Mapping.set(key, instance)

        # check all
        for key, instance in columns.items():
            mapping = self.Mapping.get(model, key)
            self.assertEqual(mapping, instance)

        # delete all
        self.Mapping.multi_delete(model, *columns.keys())

        # check all
        for key, instance in columns.items():
            mapping = self.Mapping.get(model, key)
            self.assertNotEqual(mapping, instance)
            self.assertEqual(mapping, None)

    def test_multi_set_the_same_key_with_raise(self):
        column = self.Column.query().first()
        self.Mapping.set('test_set', column)
        with self.assertRaises(IOMappingSetException):
            self.Mapping.set('test_set', column)

    def test_multi_set_the_same_key_without_raise(self):
        column = self.Column.query().first()
        self.Mapping.set('test_set', column)
        self.Mapping.set('test_set', column, raiseifexist=False)

    def test_clean_all(self):
        blok = self.Blok.insert(name='Test', version='0.0.0')
        self.Mapping.set('test', blok)
        self.registry.execute("DELETE FROM system_blok WHERE name='Test'")
        self.assertTrue(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())
        removed = self.Mapping.clean()
        self.assertEqual(removed, 1)
        self.assertFalse(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())

    def test_clean_by_bloknames(self):
        self.Blok.insert(name='Test', version='0.0.0')
        blok = self.Blok.insert(name='Test2', version='0.0.0')
        self.Mapping.set('test', blok, blokname='Test')
        self.registry.execute("DELETE FROM system_blok WHERE name='Test2'")
        self.assertTrue(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())
        removed = self.Mapping.clean(bloknames=['wrong'])
        self.assertEqual(removed, 0)
        self.assertTrue(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())
        removed = self.Mapping.clean(bloknames=['Test'])
        self.assertEqual(removed, 1)
        self.assertFalse(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())

    def test_clean_by_blokname(self):
        self.Blok.insert(name='Test', version='0.0.0')
        blok = self.Blok.insert(name='Test2', version='0.0.0')
        self.Mapping.set('test', blok, blokname='Test')
        self.registry.execute("DELETE FROM system_blok WHERE name='Test2'")
        self.assertTrue(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())
        self.Mapping.clean(bloknames='Test')
        self.assertFalse(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())

    def test_clean_by_models_1(self):
        blok = self.Blok.insert(name='Test', version='0.0.0')
        self.Mapping.set('test', blok)
        self.registry.execute("DELETE FROM system_blok WHERE name='Test'")
        self.assertTrue(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())
        self.Mapping.clean(models=['Model.System.Column'])
        self.assertTrue(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())
        self.Mapping.clean(models=['Model.System.Blok'])
        self.assertFalse(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())

    def test_clean_by_models_2(self):
        blok = self.Blok.insert(name='Test', version='0.0.0')
        self.Mapping.set('test', blok)
        self.registry.execute("DELETE FROM system_blok WHERE name='Test'")
        self.assertTrue(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())
        self.Mapping.clean(models=[self.Column])
        self.assertTrue(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())
        self.Mapping.clean(models=[self.Blok])
        self.assertFalse(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())

    def test_clean_by_model(self):
        blok = self.Blok.insert(name='Test', version='0.0.0')
        self.Mapping.set('test', blok)
        self.registry.execute("DELETE FROM system_blok WHERE name='Test'")
        self.assertTrue(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())
        self.Mapping.clean(models=self.Blok)
        self.assertFalse(self.Mapping.query().filter_by(
            key='test', model='Model.System.Blok').count())

    def test_detect_key_from_model_and_primary_key(self):
        Mapping = self.registry.IO.Mapping
        Blok = self.registry.System.Blok

        values = []
        for i, blok in enumerate(Blok.query().all()):
            values.append(dict(key='key_%d' % i, model=Blok.__registry_name__,
                               primary_key=blok.to_primary_keys()))

        Mapping.multi_insert(*values)
        mapping = Mapping.get_from_model_and_primary_keys(
            Blok.__registry_name__, blok.to_primary_keys())
        entry = Mapping.get(blok.__registry_name__, mapping.key)
        self.assertEqual(entry, blok)

    def test_delete_for_blokname(self):
        self.Blok.insert(name='Test', version='0.0.0')
        self.assertFalse(
            self.Mapping.query().filter_by(blokname="Test").count()
        )
        for column in self.Column.query().limit(10).all():
            self.Mapping.set('test_' + column.code, column, blokname='Test')
        self.assertTrue(
            self.Mapping.query().filter_by(blokname="Test").count()
        )
        removed = self.Mapping.delete_for_blokname('Test')
        self.assertEqual(removed, 10)
        self.assertFalse(
            self.Mapping.query().filter_by(blokname="Test").count()
        )

    def test_delete_for_blokname_filter_by_model_1(self):
        self.Blok.insert(name='Test', version='0.0.0')
        nb_column = self.Column.query().count()
        self.assertFalse(
            self.Mapping.query().filter_by(blokname="Test").count()
        )
        for field in self.Field.query().all():
            self.Mapping.set('test_' + field.code, field, blokname='Test')

        self.assertEqual(
            self.Mapping.query().filter_by(blokname="Test").count(),
            self.Field.query().count()
        )
        removed = self.Mapping.delete_for_blokname(
            'Test', ['Model.System.Column'])
        self.assertEqual(removed, nb_column)

    def test_delete_for_blokname_filter_by_model_2(self):
        nb_column = self.Column.query().count()
        self.Blok.insert(name='Test', version='0.0.0')
        self.assertFalse(
            self.Mapping.query().filter_by(blokname="Test").count()
        )
        for field in self.Field.query().all():
            self.Mapping.set('test_' + field.code, field, blokname='Test')

        self.assertEqual(
            self.Mapping.query().filter_by(blokname="Test").count(),
            self.Field.query().count()
        )
        removed = self.Mapping.delete_for_blokname(
            'Test', [self.registry.System.Column])
        self.assertEqual(removed, nb_column)

    def test_delete_for_blokname_filter_by_model_3(self):
        nb_column = self.Column.query().count()
        self.Blok.insert(name='Test', version='0.0.0')
        self.assertFalse(
            self.Mapping.query().filter_by(blokname="Test").count()
        )
        for field in self.Field.query().all():
            self.Mapping.set('test_' + field.code, field, blokname='Test')

        self.assertEqual(
            self.Mapping.query().filter_by(blokname="Test").count(),
            self.Field.query().count()
        )
        removed = self.Mapping.delete_for_blokname(
            'Test', self.registry.System.Column)
        self.assertEqual(removed, nb_column)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.common import anyblok_column_prefix
from sqlalchemy.orm import query


@Declarations.register(Declarations.Core)
class Query(query.Query):
    """ Overload the SqlAlchemy Query
    """

    def all(self):
        """ Return an instrumented list of the result of the query
        """
        return self.registry.InstrumentedList(super(Query, self).all())

    def with_perm(self, principals, permission):
        """Add authorization pre- and post-filtering to query.

        This must be last in the construction chain of the query.
        Queries too complicated for the authorization system to infer
        safely will be refused.

        :param principals: list, set or tuple of strings
        :param str permission: the permission to filter for
        :returns: a query-like object, with only the returning methods, such
                  as ``all()``, ``count()`` etc. available.
        """
        return self.registry.wrap_query_permission(
            self, principals, permission)

    def get_field_nams_in_column_description(self):
        field2get = [x['name'] for x in self.column_descriptions
                     if not hasattr(x['type'], '__table__')]
        field2get = [(x[len(anyblok_column_prefix):]
                      if x.startswith(anyblok_column_prefix)
                      else x, x)
                     for x in field2get]
        return field2get

    def dictone(self):
        val = self.one()
        field2get = self.get_field_nams_in_column_description()
        if field2get:
            return {x: getattr(val, y) for x, y in field2get}
        else:
            return val.to_dict()

    def dictfirst(self):
        val = self.first()
        field2get = self.get_field_nams_in_column_description()
        if field2get:
            return {x: getattr(val, y) for x, y in field2get}
        else:
            return val.to_dict()

    def dictall(self):
        vals = self.all()
        if not vals:
            return []

        field2get = self.get_field_nams_in_column_description()
        if field2get:
            return [{x: getattr(y, z) for x, z in field2get} for y in vals]
        else:
            return vals.to_dict()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations, reload_module_if_blok_is_reloading


@Declarations.register(Declarations.Model)
class System:
    pass

from . import model  # noqa
reload_module_if_blok_is_reloading(model)
from . import field  # noqa
reload_module_if_blok_is_reloading(field)
from . import column  # noqa
reload_module_if_blok_is_reloading(column)
from . import relationship  # noqa
reload_module_if_blok_is_reloading(relationship)
from . import blok  # noqa
reload_module_if_blok_is_reloading(blok)
from . import cache  # noqa
reload_module_if_blok_is_reloading(cache)
from . import parameter  # noqa
reload_module_if_blok_is_reloading(parameter)
from . import sequence  # noqa
reload_module_if_blok_is_reloading(sequence)
